from abc import ABC

from ...mixins import ErrorResponseMixin
from ...response import ControlResponse


class DeleteIndexResponse(ControlResponse):
    """Parent response type for a delete index request.

    The response object is resolved to a type-safe object of one of
    the following subtypes:
    - `DeleteIndex.Success`
    - `DeleteIndex.Error`

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+::

        match response:
            case DeleteIndex.Success():
                ...
            case DeleteIndex.Error():
                ...
            case _:
                # Shouldn't happen

    or equivalently in earlier versions of python::

        if isinstance(response, DeleteIndex.Success):
            ...
        elif isinstance(response, DeleteIndex.Error):
            ...
        else:
            # Shouldn't happen
    """


class DeleteIndex(ABC):
    """Groups all `DeleteIndexResponse` derived types under a common namespace."""

    class Success(DeleteIndexResponse):
        """Indicates the request was successful."""

    class Error(DeleteIndexResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from momento_wire_types import controlclient_pb2 as ctrl_pb

from momento.errors.exceptions import UnknownException
from momento.requests.vector_index import SimilarityMetric

from ...mixins import ErrorResponseMixin
from ...response import ControlResponse


class ListIndexesResponse(ControlResponse):
    """Parent response type for a list indexes request.

    The response object is resolved to a type-safe object of one of
    the following subtypes:
    - `ListIndexes.Success`
    - `ListIndexes.Error`

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+::

        match response:
            case ListIndexes.Success():
                ...
            case ListIndexes.Error():
                ...
            case _:
                # Shouldn't happen

    or equivalently in earlier versions of python::

        if isinstance(response, ListIndexes.Success):
            ...
        elif isinstance(response, ListIndexes.Error):
            ...
        else:
            # Shouldn't happen
    """


@dataclass
class IndexInfo:
    """Contains a Momento index's info."""

    name: str
    num_dimensions: int
    similarity_metric: SimilarityMetric

    @staticmethod
    def from_grpc_response(grpc_index_info: ctrl_pb._ListIndexesResponse._Index) -> "IndexInfo":
        metric_type: str = grpc_index_info.similarity_metric.WhichOneof("similarity_metric")
        similarity_metric: SimilarityMetric

        if metric_type == "cosine_similarity":
            similarity_metric = SimilarityMetric.COSINE_SIMILARITY
        elif metric_type == "euclidean_similarity":
            similarity_metric = SimilarityMetric.EUCLIDEAN_SIMILARITY
        elif metric_type == "inner_product":
            similarity_metric = SimilarityMetric.INNER_PRODUCT
        else:
            raise UnknownException(f"Unknown similarity metric: {metric_type}")

        return IndexInfo(
            name=grpc_index_info.index_name,
            num_dimensions=grpc_index_info.num_dimensions,
            similarity_metric=similarity_metric,
        )


class ListIndexes(ABC):
    """Groups all `ListIndexesResponse` derived types under a common namespace."""

    @dataclass
    class Success(ListIndexesResponse):
        """Indicates the request was successful."""

        indexes: list[IndexInfo]
        """The list of indexes available to the user."""

        @staticmethod
        def from_grpc_response(grpc_list_index_response: ctrl_pb._ListIndexesResponse) -> ListIndexes.Success:
            """Initializes ListIndexResponse to handle list index response.

            Args:
                grpc_list_index_response: Protobuf based response returned by Scs.
            """
            return ListIndexes.Success(
                indexes=[IndexInfo.from_grpc_response(index) for index in grpc_list_index_response.indexes]  # type: ignore[misc]  # noqa: E501
            )

    class Error(ListIndexesResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSortedSetFetchResponse(CacheResponse):
    """Response type for a `sorted_set_fetch` request.

    Its subtypes are:
    - `CacheSortedSetFetch.Hit`
    - `CacheSortedSetFetch.Miss`
    - `CacheSortedSetFetch.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetFetch(ABC):
    """Groups all `CacheSortedSetFetchResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheSortedSetFetchResponse):
        """Indicates the sorted set exists and its values were fetched."""

        value_list_bytes: list[tuple[bytes, float]]
        """The values for the fetched sorted set, as a list of bytes values to float score tuples."""

        @property
        def value_list_string(self) -> list[tuple[str, float]]:
            """The values for the fetched sorted set, as a list of utf-8 encoded string values to float score tuples.

            Returns:
                list[tuple[str, float]]
            """
            return [(key.decode("utf-8"), value) for key, value in self.value_list_bytes]

    class Miss(CacheSortedSetFetchResponse):
        """Indicates the sorted set does not exist."""

    class Error(CacheSortedSetFetchResponse, ErrorResponseMixin):
        """Indicates an error occurred in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from enum import Enum


class Service(Enum):
    CACHE = "cache"
    TOPICS = "topics"
    INDEX = "index"
    AUTH = "auth"  # not really a service we provide, but we have APIs for auth that applies to all services



from __future__ import annotations

from datetime import timedelta
from typing import Optional

from momento_wire_types import vectorindex_pb2 as vectorindex_pb
from momento_wire_types import vectorindex_pb2_grpc as vectorindex_grpc

from momento import logs
from momento.auth import CredentialProvider
from momento.config import VectorIndexConfiguration
from momento.errors import convert_error
from momento.internal._utilities import _validate_index_name, _validate_top_k
from momento.internal.services import Service
from momento.internal.synchronous._vector_index_grpc_manager import _VectorIndexDataGrpcManager
from momento.requests.vector_index import AllMetadata, FilterExpression, Item
from momento.responses.vector_index import (
    CountItems,
    CountItemsResponse,
    DeleteItemBatch,
    DeleteItemBatchResponse,
    GetItemBatch,
    GetItemBatchResponse,
    GetItemMetadataBatch,
    GetItemMetadataBatchResponse,
    Search,
    SearchAndFetchVectors,
    SearchAndFetchVectorsHit,
    SearchAndFetchVectorsResponse,
    SearchHit,
    SearchResponse,
    UpsertItemBatch,
    UpsertItemBatchResponse,
)


class _VectorIndexDataClient:
    """Internal vector index data client."""

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        endpoint = credential_provider.vector_endpoint
        self._logger = logs.logger
        self._logger.debug("Vector index data client instantiated with endpoint: %s", endpoint)
        self._endpoint = endpoint

        default_deadline: timedelta = configuration.get_transport_strategy().get_grpc_configuration().get_deadline()
        self._default_deadline_seconds = int(default_deadline.total_seconds())

        self._grpc_manager = _VectorIndexDataGrpcManager(configuration, credential_provider)

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def count_items(
        self,
        index_name: str,
    ) -> CountItemsResponse:
        try:
            self._log_issuing_request("CountItems", {"index_name": index_name})
            _validate_index_name(index_name)

            request = vectorindex_pb._CountItemsRequest(
                index_name=index_name, all=vectorindex_pb._CountItemsRequest.All()
            )
            response: vectorindex_pb._CountItemsResponse = self._build_stub().CountItems(
                request, timeout=self._default_deadline_seconds
            )

            self._log_received_response("CountItems", {"index_name": index_name})
            return CountItems.Success(item_count=response.item_count)
        except Exception as e:
            self._log_request_error("count_items", e)
            return CountItems.Error(convert_error(e, Service.INDEX))

    def upsert_item_batch(
        self,
        index_name: str,
        items: list[Item],
    ) -> UpsertItemBatchResponse:
        try:
            self._log_issuing_request("UpsertItemBatch", {"index_name": index_name})
            _validate_index_name(index_name)
            request = vectorindex_pb._UpsertItemBatchRequest(
                index_name=index_name,
                items=[item.to_proto() for item in items],
            )

            self._build_stub().UpsertItemBatch(request, timeout=self._default_deadline_seconds)

            self._log_received_response("UpsertItemBatch", {"index_name": index_name})
            return UpsertItemBatch.Success()
        except Exception as e:
            self._log_request_error("set", e)
            return UpsertItemBatch.Error(convert_error(e, Service.INDEX))

    def delete_item_batch(
        self,
        index_name: str,
        ids: list[str],
    ) -> DeleteItemBatchResponse:
        try:
            self._log_issuing_request("DeleteItemBatch", {"index_name": index_name})
            _validate_index_name(index_name)

            if len(ids) == 0:
                return DeleteItemBatch.Success()

            request = vectorindex_pb._DeleteItemBatchRequest(
                index_name=index_name,
                ids=ids,
            )

            self._build_stub().DeleteItemBatch(request, timeout=self._default_deadline_seconds)

            self._log_received_response("DeleteItemBatch", {"index_name": index_name})
            return DeleteItemBatch.Success()
        except Exception as e:
            self._log_request_error("delete", e)
            return DeleteItemBatch.Error(convert_error(e, Service.INDEX))

    @staticmethod
    def __build_metadata_request(metadata_fields: Optional[list[str]] | AllMetadata) -> vectorindex_pb._MetadataRequest:
        if isinstance(metadata_fields, AllMetadata):
            return vectorindex_pb._MetadataRequest(all=vectorindex_pb._MetadataRequest.All())
        else:
            return vectorindex_pb._MetadataRequest(
                some=vectorindex_pb._MetadataRequest.Some(fields=metadata_fields if metadata_fields else [])
            )

    @staticmethod
    def __build_no_score_threshold(score_threshold: Optional[float]) -> Optional[vectorindex_pb._NoScoreThreshold]:
        if score_threshold is None:
            return vectorindex_pb._NoScoreThreshold()
        return None

    @staticmethod
    def __build_filter_expression(
        filter_expression: Optional[FilterExpression],
    ) -> Optional[vectorindex_pb._FilterExpression]:
        if filter_expression is None:
            return None
        return filter_expression.to_filter_expression_proto()

    def search(
        self,
        index_name: str,
        query_vector: list[float],
        top_k: int,
        metadata_fields: Optional[list[str]] | AllMetadata = None,
        score_threshold: Optional[float] = None,
        filter: Optional[FilterExpression] = None,
    ) -> SearchResponse:
        try:
            self._log_issuing_request("Search", {"index_name": index_name})
            _validate_index_name(index_name)
            _validate_top_k(top_k)

            query_vector_pb = vectorindex_pb._Vector(elements=query_vector)
            metadata_fields_pb = _VectorIndexDataClient.__build_metadata_request(metadata_fields)
            no_score_threshold = _VectorIndexDataClient.__build_no_score_threshold(score_threshold)
            filter_expression_pb = _VectorIndexDataClient.__build_filter_expression(filter)

            request = vectorindex_pb._SearchRequest(
                index_name=index_name,
                query_vector=query_vector_pb,
                top_k=top_k,
                metadata_fields=metadata_fields_pb,
                score_threshold=score_threshold,
                no_score_threshold=no_score_threshold,
                filter=filter_expression_pb,
            )

            response: vectorindex_pb._SearchResponse = self._build_stub().Search(
                request, timeout=self._default_deadline_seconds
            )

            hits = [SearchHit.from_proto(hit) for hit in response.hits]
            self._log_received_response("Search", {"index_name": index_name})
            return Search.Success(hits=hits)
        except Exception as e:
            self._log_request_error("search", e)
            return Search.Error(convert_error(e, Service.INDEX))

    def search_and_fetch_vectors(
        self,
        index_name: str,
        query_vector: list[float],
        top_k: int,
        metadata_fields: Optional[list[str]] | AllMetadata = None,
        score_threshold: Optional[float] = None,
        filter: Optional[FilterExpression] = None,
    ) -> SearchAndFetchVectorsResponse:
        try:
            self._log_issuing_request("SearchAndFetchVectors", {"index_name": index_name})
            _validate_index_name(index_name)
            _validate_top_k(top_k)

            query_vector_pb = vectorindex_pb._Vector(elements=query_vector)
            metadata_fields_pb = _VectorIndexDataClient.__build_metadata_request(metadata_fields)
            no_score_threshold = _VectorIndexDataClient.__build_no_score_threshold(score_threshold)
            filter_expression_pb = _VectorIndexDataClient.__build_filter_expression(filter)

            request = vectorindex_pb._SearchAndFetchVectorsRequest(
                index_name=index_name,
                query_vector=query_vector_pb,
                top_k=top_k,
                metadata_fields=metadata_fields_pb,
                score_threshold=score_threshold,
                no_score_threshold=no_score_threshold,
                filter=filter_expression_pb,
            )

            response: vectorindex_pb._SearchAndFetchVectorsResponse = self._build_stub().SearchAndFetchVectors(
                request, timeout=self._default_deadline_seconds
            )

            hits = [SearchAndFetchVectorsHit.from_proto(hit) for hit in response.hits]
            self._log_received_response("SearchAndFetchVectors", {"index_name": index_name})
            return SearchAndFetchVectors.Success(hits=hits)
        except Exception as e:
            self._log_request_error("search_and_fetch_vectors", e)
            return SearchAndFetchVectors.Error(convert_error(e, Service.INDEX))

    def get_item_batch(
        self,
        index_name: str,
        ids: list[str],
    ) -> GetItemBatchResponse:
        try:
            self._log_issuing_request("GetItemBatch", {"index_name": index_name})
            _validate_index_name(index_name)

            if len(ids) == 0:
                return GetItemBatch.Success(values={})

            request = vectorindex_pb._GetItemBatchRequest(
                index_name=index_name,
                ids=ids,
                metadata_fields=vectorindex_pb._MetadataRequest(all=vectorindex_pb._MetadataRequest.All()),
            )

            batch_response: vectorindex_pb._GetItemBatchResponse = self._build_stub().GetItemBatch(
                request, timeout=self._default_deadline_seconds
            )
            self._log_received_response("GetItemBatch", {"index_name": index_name})
            return GetItemBatch.Success.from_proto(batch_response)
        except Exception as e:
            self._log_request_error("get_item_batch", e)
            return GetItemBatch.Error(convert_error(e, Service.INDEX))

    def get_item_metadata_batch(
        self,
        index_name: str,
        ids: list[str],
    ) -> GetItemMetadataBatchResponse:
        try:
            self._log_issuing_request("GetItemMetadataBatch", {"index_name": index_name})
            _validate_index_name(index_name)

            if len(ids) == 0:
                return GetItemMetadataBatch.Success(values={})

            request = vectorindex_pb._GetItemMetadataBatchRequest(
                index_name=index_name,
                ids=ids,
                metadata_fields=vectorindex_pb._MetadataRequest(all=vectorindex_pb._MetadataRequest.All()),
            )

            batch_response: vectorindex_pb._GetItemMetadataBatchResponse = self._build_stub().GetItemMetadataBatch(
                request, timeout=self._default_deadline_seconds
            )
            self._log_received_response("GetItemMetadataBatch", {"index_name": index_name})
            return GetItemMetadataBatch.Success.from_proto(batch_response)
        except Exception as e:
            self._log_request_error("get_item_metadata_batch", e)
            return GetItemMetadataBatch.Error(convert_error(e, Service.INDEX))

    # TODO these were copied from the data client. Shouldn't use interpolation here for perf?
    def _log_received_response(self, request_type: str, request_args: dict[str, str]) -> None:
        self._logger.log(logs.TRACE, f"Received a {request_type} response for {request_args}")

    def _log_issuing_request(self, request_type: str, request_args: dict[str, str]) -> None:
        self._logger.log(logs.TRACE, f"Issuing a {request_type} request with {request_args}")

    def _log_request_error(self, request_type: str, e: Exception) -> None:
        self._logger.warning(f"{request_type} failed with exception: {e}")

    def _build_stub(self) -> vectorindex_grpc.VectorIndexStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



from __future__ import annotations

from typing import Iterable

from momento_wire_types import vectorindex_pb2 as pb

from momento.common_data.vector_index.item import Metadata
from momento.errors import UnknownException


def pb_metadata_to_dict(
    pb_metadata: Iterable[pb._Metadata],
) -> Metadata:
    metadata: Metadata = {}
    for item in pb_metadata:
        type = item.WhichOneof("value")
        field = item.field
        if type == "string_value":
            metadata[field] = item.string_value
        elif type == "integer_value":
            metadata[field] = item.integer_value
        elif type == "double_value":
            metadata[field] = item.double_value
        elif type == "boolean_value":
            metadata[field] = item.boolean_value
        elif type == "list_of_strings_value":
            metadata[field] = list(item.list_of_strings_value.values)
        else:
            raise UnknownException(f"Unknown metadata value: {type}")
    return metadata



from __future__ import annotations

import grpc
from momento_wire_types import controlclient_pb2_grpc as control_client
from momento_wire_types import vectorindex_pb2_grpc as vector_index_client

from momento.auth import CredentialProvider
from momento.config import VectorIndexConfiguration
from momento.internal._utilities import momento_version
from momento.internal._utilities._channel_credentials import (
    channel_credentials_from_root_certs_or_default,
)
from momento.internal.synchronous._add_header_client_interceptor import (
    AddHeaderClientInterceptor,
    Header,
)


class _VectorIndexControlGrpcManager:
    """Internal gRPC control mananger."""

    version = momento_version

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.control_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
        )
        intercept_channel = grpc.intercept_channel(self._secure_channel, *_interceptors(credential_provider.auth_token))
        self._stub = control_client.ScsControlStub(intercept_channel)  # type: ignore[no-untyped-call]

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> control_client.ScsControlStub:
        return self._stub


class _VectorIndexDataGrpcManager:
    """Internal gRPC vector index data manager."""

    version = momento_version

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.secure_channel(
            target=credential_provider.vector_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
        )
        intercept_channel = grpc.intercept_channel(self._secure_channel, *_interceptors(credential_provider.auth_token))
        self._stub = vector_index_client.VectorIndexStub(intercept_channel)  # type: ignore[no-untyped-call]

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> vector_index_client.VectorIndexStub:
        return self._stub


def _interceptors(auth_token: str) -> list[grpc.UnaryUnaryClientInterceptor]:
    headers = [Header("authorization", auth_token), Header("agent", f"python:{_VectorIndexControlGrpcManager.version}")]
    return list(filter(None, [AddHeaderClientInterceptor(headers)]))



from abc import ABC

from ..mixins import ErrorResponseMixin
from ..response import PubsubResponse


class TopicPublishResponse(PubsubResponse):
    """Parent response type for a topic `publish` request.

    Its subtypes are:
    - `TopicPublish.Success`
    - `TopicPublish.Error`
    """


class TopicPublish(ABC):
    """Groups all `TopicPublishResponse` derived types under a common namespace."""

    class Success(TopicPublishResponse):
        """Indicates the request was successful."""

    class Error(TopicPublishResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse
from .get_score import CacheSortedSetGetScore, CacheSortedSetGetScoreResponse


class CacheSortedSetGetScoresResponse(CacheResponse):
    """Parent response type for a cache `sorted_set_get_scores` request.

    Its subtypes are:
    - `CacheSortedSetGetScores.Hit`
    - `CacheSortedSetGetScores.Miss`
    - `CacheSortedSetGetScores.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetGetScores(ABC):
    """Groups all `CacheSortedSetGetScoresResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheSortedSetGetScoresResponse):
        """Contains the result of a cache hit."""

        responses: list[CacheSortedSetGetScoreResponse]
        """The value returned from the cache for the specified field. Use the
        `value_string` property to access the value as a string."""

        @property
        def value_dictionary_bytes(self) -> dict[bytes, float]:
            """The values for the fetched sorted set, as bytes to their float scores.

            Returns:
                dict[bytes, float]
            """
            return {
                response.value_bytes: response.score
                for response in self.responses
                if isinstance(response, CacheSortedSetGetScore.Hit)
            }

        @property
        def value_dictionary_string(self) -> dict[str, float]:
            """The values for the fetched sorted set, as utf-8 encoded strings to their float scores.

            Returns:
                dict[str, float]
            """
            return {
                response.value_string: response.score
                for response in self.responses
                if isinstance(response, CacheSortedSetGetScore.Hit)
            }

    class Miss(CacheSortedSetGetScoresResponse):
        """Indicates the sorted set does not exist."""

    class Error(CacheSortedSetGetScoresResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from momento_wire_types import vectorindex_pb2 as pb

from momento.common_data.vector_index.item import Item
from momento.errors.exceptions import UnknownException

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse
from .utils import pb_metadata_to_dict


class GetItemBatchResponse(VectorIndexResponse):
    """Parent response type for a `get_item_batch` request.

    Its subtypes are:
    - `GetItemBatch.Success`
    - `GetItemBatch.Error`

    See `PreviewVectorIndexClient` for how to work with responses.
    """


class GetItemBatch(ABC):
    """Groups all `GetItemBatchResponse` derived types under a common namespace."""

    @dataclass
    class Success(GetItemBatchResponse):
        """Contains the result of a `get_item_batch` request."""

        values: dict[str, Item]
        """The items that were found."""

        @staticmethod
        def from_proto(response: pb._GetItemBatchResponse) -> "GetItemBatch.Success":
            """Converts a proto hit to a `GetItemBatch.Success`."""
            values = {}
            for item in response.item_response:
                type = item.WhichOneof("response")
                if type == "hit":
                    id_, metadata = item.hit.id, pb_metadata_to_dict(item.hit.metadata)
                    values[id_] = Item(id=id_, vector=list(item.hit.vector.elements), metadata=metadata)
                elif type == "miss":
                    pass
                else:
                    raise UnknownException(f"Unknown response type {type!r}.")

            return GetItemBatch.Success(values=values)

    class Error(GetItemBatchResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from pathlib import Path
from typing import Optional


class GrpcConfiguration(ABC):
    @abstractmethod
    def get_deadline(self) -> timedelta:
        pass

    @abstractmethod
    def with_deadline(self, deadline: timedelta) -> GrpcConfiguration:
        pass

    @abstractmethod
    def with_root_certificates_pem(self, root_certificates_pem_path: Path) -> GrpcConfiguration:
        pass

    @abstractmethod
    def get_root_certificates_pem(self) -> Optional[bytes]:
        pass



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSortedSetIncrementScoreResponse(CacheResponse):
    """Parent response type for a cache `sorted_set_increment_score` request.

    Its subtypes are:
    - `CacheSortedSetIncrementScore.Success`
    - `CacheSortedSetIncrementScore.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetIncrementScore(ABC):
    """Groups all `CacheSortedSetIncrementScoreResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheSortedSetIncrementScoreResponse):
        """Indicates the request was successful."""

        score: float
        """The score of the element post-increment."""

    class Error(CacheSortedSetIncrementScoreResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSortedSetRemoveElementsResponse(CacheResponse):
    """Response type for a `sorted_set_remove_elements` request.

    Its subtypes are:
    - `CacheSortedSetRemoveElements.Success`
    - `CacheSortedSetRemoveElements.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetRemoveElements(ABC):
    """Groups all `CacheSortedSetRemoveElementsResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheSortedSetRemoveElementsResponse):
        """Indicates the remove was successful."""

    class Error(CacheSortedSetRemoveElementsResponse, ErrorResponseMixin):
        """Indicates an error occurred in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from concurrent.futures._base import CancelledError
from typing import Optional

from grpc._channel import _MultiThreadedRendezvous
from grpc.aio._interceptor import InterceptedUnaryStreamCall
from momento_wire_types import cachepubsub_pb2

from ... import logs
from ...errors import MomentoErrorCode, SdkException
from ...internal.services import Service
from ..mixins import ErrorResponseMixin
from ..response import PubsubResponse
from .subscription_item import TopicSubscriptionItem, TopicSubscriptionItemResponse


class TopicSubscribeResponse(PubsubResponse):
    """Parent response type for a topic `publish` request.

    Its subtypes are:
    - `TopicSubscribe.Subscription`
    - `TopicSubscribe.SubscriptionAsync`
    - `TopicSubscribe.Error`

    Both the ``Subscription` and the `SubscriptionAsync` response subtypes are
    iterators that yield subscription items.
    """


class TopicSubscribe(ABC):
    """Groups all `TopicSubscribeResponse` derived types under a common namespace."""

    class SubscriptionBase(TopicSubscribeResponse):
        """Base class for common logic shared between async and synchronous subscriptions."""

        _logger = logs.logger
        _last_known_sequence_number: Optional[int] = None

        def _process_result(self, result: cachepubsub_pb2._SubscriptionItem) -> Optional[TopicSubscriptionItemResponse]:
            msg_type: str = result.WhichOneof("kind")
            if msg_type == "item":
                self._last_known_sequence_number = result.item.topic_sequence_number
                value = result.item.value
                value_type: str = value.WhichOneof("kind")
                if value_type == "text":
                    return TopicSubscriptionItem.Text(value.text)
                elif value_type == "binary":
                    return TopicSubscriptionItem.Binary(value.binary)
                else:
                    err = SdkException(
                        f"Could not find matching TopicSubscriptionItem response for type: {value_type}",
                        MomentoErrorCode.UNKNOWN_ERROR,
                        Service.TOPICS,
                        message_wrapper="Error reading item from topic subscription",
                    )
                    return TopicSubscriptionItem.Error(err)
            elif msg_type == "heartbeat":
                self._logger.debug("client stream received heartbeat")
                return None
            elif msg_type == "discontinuity":
                self._logger.debug("client stream received discontinuity")
                return None
            self._logger.debug(f"client stream received unknown type: {msg_type}")
            return None

    class SubscriptionAsync(SubscriptionBase):
        """Provides the async version of a topic subscription."""

        def __init__(self, cache_name: str, topic_name: str, client_stream: InterceptedUnaryStreamCall):
            self._cache_name = cache_name
            self._topic_name = topic_name
            self._client_stream = client_stream  # type: ignore[misc]

        def __aiter__(self) -> TopicSubscribe.SubscriptionAsync:
            return self

        async def __anext__(self) -> TopicSubscriptionItemResponse:
            """Retrieves the next published item from the subscription.

            The item method returns a response object that is resolved to a type-safe object
            of one of the following:

            - TopicSubscriptionItem.Text
            - TopicSubscriptionItem.Binary
            - TopicSubscriptionItem.Error

            Pattern matching can be used to operate on the appropriate subtype.
            For example, in python 3.10+ if you're receiving a subscription item::

                async for response in subscription:
                    match response:
                        case TopicSubscriptionItem.Text():
                            return response.value
                        case TopicSubscriptionItem.Binary():
                            return response.value
                        case TopicSubscriptionItem.Error():
                            ...there was an error retrieving the item...

            or equivalently in earlier versions of python::

                async for response in subscription:
                    if isinstance(response, TopicSubscriptionItem.Text):
                        return response.value
                    elif isinstance(response, TopicSubscriptionItem.Binary):
                        return response.value
                    elif isinstance(response, TopicSubscriptionItem.Error):
                        ...there was an error retrieving the item...
            """
            while True:
                try:
                    result: cachepubsub_pb2._SubscriptionItem = await self._client_stream.read()  # type: ignore[misc]
                    item = self._process_result(result)
                    # if item is null, we've received a heartbeat or discontinuity and should continue
                    if item is not None:
                        return item
                except (CancelledError, StopAsyncIteration) as e:
                    err = SdkException(
                        f"Client subscription has been cancelled by {type(e)}",
                        MomentoErrorCode.CANCELLED_ERROR,
                        Service.TOPICS,
                        message_wrapper="Error reading item from topic subscription",
                    )
                    self._logger.debug(f"Client stream read has been cancelled: {type(e)}")
                    return TopicSubscriptionItem.Error(err)
                except Exception as e:
                    self._logger.debug(f"Error reading from client stream: {type(e)}")
                    # TODO: attempt reconnect
                    continue

    class Subscription(SubscriptionBase):
        """Provides the synchronous version of a topic subscription."""

        def __init__(self, cache_name: str, topic_name: str, client_stream: _MultiThreadedRendezvous):
            self._cache_name = cache_name
            self._topic_name = topic_name
            self._client_stream = client_stream  # type: ignore[misc]

        def __iter__(self) -> TopicSubscribe.Subscription:
            return self

        def __next__(self) -> TopicSubscriptionItemResponse:
            """Retrieves the next published item from the subscription.

            The item method returns a response object that is resolved to a type-safe object
            of one of the following:

            - TopicSubscriptionItem.Text
            - TopicSubscriptionItem.Binary
            - TopicSubscriptionItem.Error

            Pattern matching can be used to operate on the appropriate subtype.
            For example, in python 3.10+ if you're receiving a subscription item::

                for response in subscription:
                    match response:
                        case TopicSubscriptionItem.Text():
                            return response.value
                        case TopicSubscriptionItem.Binary():
                            return response.value
                        case TopicSubscriptionItem.Error():
                            ...there was an error retrieving the item...

            or equivalently in earlier versions of python::

                for response in subscription:
                    if isinstance(response, TopicSubscriptionItem.Text):
                        return response.value
                    elif isinstance(response, TopicSubscriptionItem.Binary):
                        return response.value
                    elif isinstance(response, TopicSubscriptionItem.Error):
                        ...there was an error retrieving the item...
            """
            while True:
                try:
                    result: cachepubsub_pb2._SubscriptionItem = self._client_stream.next()  # type: ignore[misc]
                    item = self._process_result(result)
                    # if item is null, we've received a heartbeat or discontinuity and should continue
                    if item is not None:
                        return item
                except Exception as e:
                    self._logger.debug("Error reading from client stream: %s", e)
                    # TODO: attempt reconnect
                    continue

    class Error(TopicSubscribeResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from datetime import timedelta

DEFAULT_EAGER_CONNECTION_TIMEOUT_SECONDS = 30


def validate_eager_connection_timeout(timeout: timedelta) -> None:
    if timeout.total_seconds() < 0:
        raise ValueError("The eager connection timeout must be greater than or equal to 0 seconds.")



from __future__ import annotations

from datetime import timedelta

from .transport.transport_strategy import (
    StaticGrpcConfiguration,
    StaticTransportStrategy,
)
from .vector_index_configuration import VectorIndexConfiguration


class VectorIndexConfigurations:
    """Container class for pre-built configurations."""

    class Default(VectorIndexConfiguration):
        """Laptop config provides defaults suitable for a medium-to-high-latency dev environment.

        Permissive timeouts, retries, and relaxed latency and throughput targets.
        """

        @staticmethod
        def latest() -> VectorIndexConfigurations.Default:
            """Provides the latest recommended configuration for a laptop development environment.

            This configuration will be updated every time there is a new version of the laptop configuration.
            """
            return VectorIndexConfigurations.Default(
                # This is high to account for time-intensive adds.
                StaticTransportStrategy(StaticGrpcConfiguration(timedelta(seconds=120)))
            )



from __future__ import annotations

from typing import Dict, List, Optional, Union

from momento_wire_types import vectorindex_pb2 as pb

from momento.errors.exceptions import InvalidArgumentException
from momento.internal.services import Service

Metadata = Dict[str, Union[str, int, float, bool, List[str]]]
"""The metadata of an item."""


class ItemMetadata:
    """Represents the id-metadata portion of an entry in the vector index.

    This is used in requests to update only the metadata of an item.
    """

    def __init__(self, id: str, metadata: Optional[Metadata] = None) -> None:
        """Represents the id-metadata portion of an entry in the vector index.

        Args:
            id (str): The id of the item.
            metadata (Optional[Metadata], optional): The metadata of the item. Defaults to None, ie empty metadata.
        """
        self.id = id
        self.metadata = metadata or {}

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ItemMetadata):
            return self.id == other.id and self.metadata == other.metadata
        return False

    def __hash__(self) -> int:
        return hash((self.id, self.metadata))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id!r}, metadata={self.metadata!r})"


# TODO: support other datatypes for the vector (np.array, pd.Series, etc.)
class Item(ItemMetadata):
    """Represents a full entry in the vector index.

    This is used for `upsert_item_batch` requests.
    """

    def __init__(self, id: str, vector: list[float], metadata: Optional[Metadata] = None) -> None:
        """Represents an entry in the vector index.

        Args:
            id (str): The id of the item.
            vector (list[float]): The vector of the item.
            metadata (Optional[Metadata], optional): The metadata of the item. Defaults to None, ie empty metadata.
        """
        super().__init__(id, metadata)
        self.vector = vector

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Item):
            return super().__eq__(other) and self.vector == other.vector
        return False

    def __hash__(self) -> int:
        return hash((self.id, self.vector, self.metadata))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id!r}, vector={self.vector!r}, metadata={self.metadata!r})"

    def to_proto(self) -> pb._Item:
        vector = pb._Vector(elements=self.vector)
        metadata = []
        for k, v in self.metadata.items():
            if type(v) is str:
                metadata.append(pb._Metadata(field=k, string_value=v))
            elif type(v) is int:
                metadata.append(pb._Metadata(field=k, integer_value=v))
            elif type(v) is float:
                metadata.append(pb._Metadata(field=k, double_value=v))
            elif type(v) is bool:
                metadata.append(pb._Metadata(field=k, boolean_value=v))
            elif type(v) is list and all(type(x) is str for x in v):
                list_of_strings = pb._Metadata._ListOfStrings(values=v)
                metadata.append(pb._Metadata(field=k, list_of_strings_value=list_of_strings))
            else:
                raise InvalidArgumentException(
                    f"Metadata values must be either str, int, float, bool, or list[str]. Field {k!r} has a value of type {type(v)!r} with value {v!r}.",  # noqa: E501
                    Service.INDEX,
                )
        return pb._Item(id=self.id, vector=vector, metadata=metadata)



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin, ValueStringMixin
from ...response import CacheResponse


class CacheSortedSetGetScoreResponse(CacheResponse):
    """Parent response type for a cache `sorted_set_get_score` request.

    Its subtypes are:
    - `CacheSortedSetGetScore.Hit`
    - `CacheSortedSetGetScore.Miss`
    - `CacheSortedSetGetScore.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetGetScore(ABC):
    """Groups all `CacheSortedSetGetScoreResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheSortedSetGetScoreResponse, ValueStringMixin):
        """Contains the result of a cache hit."""

        value_bytes: bytes
        """The value for which the score was queried. Use the
        `value_string` property to access the value as a string."""

        score: float
        """The score of the sorted set value that was queried."""

    @dataclass
    class Miss(CacheSortedSetGetScoreResponse, ValueStringMixin):
        """Indicates the sorted set or sorted set element does not exist."""

        value_bytes: bytes
        """The value that did not have a score. Use the
        `value_string` property to access the value as a string."""

    class Error(CacheSortedSetGetScoreResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from pathlib import Path

from .transport.transport_strategy import TransportStrategy


class VectorIndexConfigurationBase(ABC):
    @abstractmethod
    def get_transport_strategy(self) -> TransportStrategy:
        pass

    @abstractmethod
    def with_transport_strategy(self, transport_strategy: TransportStrategy) -> VectorIndexConfigurationBase:
        pass

    @abstractmethod
    def with_client_timeout(self, client_timeout: timedelta) -> VectorIndexConfigurationBase:
        pass

    @abstractmethod
    def with_root_certificates_pem(self, root_certificate_path: Path) -> VectorIndexConfigurationBase:
        pass


class VectorIndexConfiguration(VectorIndexConfigurationBase):
    """Configuration options for Momento Vector Index Client."""

    def __init__(self, transport_strategy: TransportStrategy):
        """Instantiate a Configuration.

        Args:
            transport_strategy (TransportStrategy): Configuration options for networking with
            the Momento service.
        """
        self._transport_strategy = transport_strategy

    def get_transport_strategy(self) -> TransportStrategy:
        """Access the transport strategy.

        Returns:
            TransportStrategy: the current configuration options for wire interactions with the Momento service.
        """
        return self._transport_strategy

    def with_transport_strategy(self, transport_strategy: TransportStrategy) -> VectorIndexConfiguration:
        """Copy constructor for overriding TransportStrategy.

        Args:
            transport_strategy (TransportStrategy): the new TransportStrategy.

        Returns:
            Configuration: the new Configuration with the specified TransportStrategy.
        """
        return VectorIndexConfiguration(transport_strategy)

    def with_client_timeout(self, client_timeout: timedelta) -> VectorIndexConfiguration:
        """Copies the Configuration and sets the new client-side timeout in the copy's TransportStrategy.

        Args:
            client_timeout (timedelta): the new client-side timeout.

        Return:
            Configuration: the new Configuration.
        """
        return self.with_transport_strategy(self._transport_strategy.with_client_timeout(client_timeout))

    def with_root_certificates_pem(self, root_certificates_pem_path: Path) -> VectorIndexConfiguration:
        """Copies the Configuration and sets the new root certificates in the copy's TransportStrategy.

        Args:
            root_certificates_pem_path (Path): the new root certificates.

        Returns:
            VectorIndexConfigurationBase: the new Configuration.
        """
        grpc_configuration = self._transport_strategy.get_grpc_configuration().with_root_certificates_pem(
            root_certificates_pem_path
        )
        transport_strategy = self._transport_strategy.with_grpc_configuration(grpc_configuration)
        return self.with_transport_strategy(transport_strategy)



class AllMetadata:
    """A class to represent a search for all metadata."""

    pass


ALL_METADATA = AllMetadata()
"""A special value to pass to :meth:`VectorIndexClientAsync.search` to request all metadata fields."""



from __future__ import annotations

from types import TracebackType
from typing import Optional, Type

from momento import logs
from momento.auth import CredentialProvider
from momento.config import TopicConfiguration
from momento.internal.synchronous._scs_pubsub_client import _ScsPubsubClient
from momento.responses import TopicPublishResponse, TopicSubscribeResponse


class TopicClient:
    """Synchronous Topic Client.

    Publish and subscribe methods return a response object unique to each request.
    The response object is resolved to a type-safe object of one of several
    sub-types. See the documentation for each response type for details.

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+ if you're subscribing to a topic::

        response = client.subscribe(cache_name, topic_name)
        match response:
            case TopicSubscribe.Subscription():
                ...the subscription was successful...
            case TopicSubscribe.Error():
                ...there was an error trying to subscribe...

    or equivalently in earlier versions of python::

        response = client.subscribe(cache_name, topic_name)
        if isinstance(response, TopicSubscribe.Subscription):
            ...
        elif isinstance(response, TopicSubscribe.Error):
            ...
        else:
            raise Exception("This should never happen")
    """

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        """Instantiate a client.

        Args:
            configuration (TopicConfiguration): An object holding configuration settings for communication
                with the server.
            credential_provider: (CredentialProvider): An object holding the auth token and endpoint information.

        Example:
            from momento import CredentialProvider, TopicClient, TopicConfigurations

            configuration = TopicConfigurations.Default.v1()
            credential_provider = CredentialProvider.from_environment_variable("MOMENTO_API_KEY")
            client = TopicClient(configuration, credential_provider)
        """
        self._logger = logs.logger
        self._cache_endpoint = credential_provider.cache_endpoint
        self._pubsub_client = _ScsPubsubClient(configuration, credential_provider)

    def __enter__(self) -> TopicClient:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._pubsub_client.close()

    def publish(self, cache_name: str, topic_name: str, value: str | bytes) -> TopicPublishResponse:
        """Publishes a message to a topic.

        Args:
            cache_name (str): Name of the cache to publish to.
            topic_name (str): Name of the topic to publish to.
            value (str|bytes): The value to publish

        Returns:
            TopicPublishResponse
        """
        return self._pubsub_client.publish(cache_name, topic_name, value)

    def subscribe(self, cache_name: str, topic_name: str) -> TopicSubscribeResponse:
        """Subscribes to a topic.

        Args:
            cache_name (str): The cache to subscribe to.
            topic_name (str): The topic to subscribe to.

        Returns:
            TopicSubscribeResponse
        """
        return self._pubsub_client.subscribe(cache_name, topic_name)

    def close(self) -> None:
        self._pubsub_client.close()



from abc import ABC
from dataclasses import dataclass
from warnings import warn

from ..mixins import ErrorResponseMixin, ValueStringMixin
from ..response import PubsubResponse


class TopicSubscriptionItemResponse(PubsubResponse):
    """Parent response type for a topic subscription's `item` request.

    Its subtypes are:
    - `TopicSubscriptionItem.Text`
    - `TopicSubscriptionItem.Binary`
    - `TopicSubscriptionItem.Error`

    The subtype `TopicSubscriptionItem.Success` is deprecated.
    """


class TopicSubscriptionItem(ABC):
    """Groups all `TopicSubscriptionItemResponse` derived types under a common namespace."""

    @dataclass
    class Success(TopicSubscriptionItemResponse, ValueStringMixin):
        """Indicates the request was successful."""

        value_bytes: bytes
        """The item returned from the subscription for the specified topic. Use the
        `value_string` property to access the value as a string."""

        def __post_init__(self) -> None:
            warn("Success is a deprecated response, use Text or Binary instead", DeprecationWarning, stacklevel=2)

    @dataclass
    class Text(TopicSubscriptionItemResponse):
        """Indicates the request was successful and value will be returned as a string."""

        value: str

    @dataclass
    class Binary(TopicSubscriptionItemResponse):
        """Indicates the request was successful and value will be returned as bytes."""

        value: bytes

    class Error(TopicSubscriptionItemResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



import grpc
from momento_wire_types import controlclient_pb2 as ctrl_pb
from momento_wire_types import controlclient_pb2_grpc as ctrl_grpc

from momento import logs
from momento.auth import CredentialProvider
from momento.config import VectorIndexConfiguration
from momento.errors import InvalidArgumentException, convert_error
from momento.internal._utilities import _validate_index_name, _validate_num_dimensions
from momento.internal.services import Service
from momento.internal.synchronous._vector_index_grpc_manager import (
    _VectorIndexControlGrpcManager,
)
from momento.requests.vector_index import SimilarityMetric
from momento.responses.vector_index import (
    CreateIndex,
    CreateIndexResponse,
    DeleteIndex,
    DeleteIndexResponse,
    ListIndexes,
    ListIndexesResponse,
)

_DEADLINE_SECONDS = 60.0  # 1 minute


class _VectorIndexControlClient:
    """Momento Internal."""

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        endpoint = credential_provider.control_endpoint
        self._logger = logs.logger
        self._logger.debug("Vector index control client instantiated with endpoint: %s", endpoint)
        self._grpc_manager = _VectorIndexControlGrpcManager(configuration, credential_provider)
        self._endpoint = endpoint

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def create_index(
        self, index_name: str, num_dimensions: int, similarity_metric: SimilarityMetric
    ) -> CreateIndexResponse:
        try:
            self._logger.info(f"Creating index with name: {index_name}")
            _validate_index_name(index_name)
            _validate_num_dimensions(num_dimensions)

            if similarity_metric == SimilarityMetric.EUCLIDEAN_SIMILARITY:
                request = ctrl_pb._CreateIndexRequest(
                    index_name=index_name,
                    num_dimensions=num_dimensions,
                    similarity_metric=ctrl_pb._SimilarityMetric(
                        euclidean_similarity=ctrl_pb._SimilarityMetric._EuclideanSimilarity()
                    ),
                )
            elif similarity_metric == SimilarityMetric.INNER_PRODUCT:
                request = ctrl_pb._CreateIndexRequest(
                    index_name=index_name,
                    num_dimensions=num_dimensions,
                    similarity_metric=ctrl_pb._SimilarityMetric(
                        inner_product=ctrl_pb._SimilarityMetric._InnerProduct()
                    ),
                )
            elif similarity_metric == SimilarityMetric.COSINE_SIMILARITY:
                request = ctrl_pb._CreateIndexRequest(
                    index_name=index_name,
                    num_dimensions=num_dimensions,
                    similarity_metric=ctrl_pb._SimilarityMetric(
                        cosine_similarity=ctrl_pb._SimilarityMetric._CosineSimilarity()
                    ),
                )
            else:
                raise InvalidArgumentException(f"Invalid similarity metric `{similarity_metric}`", Service.INDEX)
            self._build_stub().CreateIndex(request, timeout=_DEADLINE_SECONDS)
        except Exception as e:
            self._logger.debug("Failed to create index: %s with exception: %s", index_name, e)
            if isinstance(e, grpc.RpcError) and e.code() == grpc.StatusCode.ALREADY_EXISTS:
                return CreateIndex.IndexAlreadyExists()
            return CreateIndex.Error(convert_error(e, Service.INDEX))
        return CreateIndex.Success()

    def delete_index(self, index_name: str) -> DeleteIndexResponse:
        try:
            self._logger.info(f"Deleting index with name: {index_name}")
            _validate_index_name(index_name)
            request = ctrl_pb._DeleteIndexRequest(index_name=index_name)
            self._build_stub().DeleteIndex(request, timeout=_DEADLINE_SECONDS)
        except Exception as e:
            self._logger.debug("Failed to delete index: %s with exception: %s", index_name, e)
            return DeleteIndex.Error(convert_error(e, Service.INDEX))
        return DeleteIndex.Success()

    def list_indexes(self) -> ListIndexesResponse:
        try:
            list_indexes_request = ctrl_pb._ListIndexesRequest()
            response = self._build_stub().ListIndexes(list_indexes_request, timeout=_DEADLINE_SECONDS)
            return ListIndexes.Success.from_grpc_response(response)
        except Exception as e:
            return ListIndexes.Error(convert_error(e, Service.INDEX))

    def _build_stub(self) -> ctrl_grpc.ScsControlStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



from __future__ import annotations

from momento.errors import InvalidArgumentException
from momento.internal.services import Service

from ._data_validation import _validate_name


def _validate_index_name(cache_name: str) -> None:
    _validate_name(cache_name, "Vector index name", Service.INDEX)


def _validate_num_dimensions(num_dimensions: int) -> None:
    if num_dimensions < 1 or not isinstance(num_dimensions, int):
        raise InvalidArgumentException("Number of dimensions must be a positive integer.", Service.INDEX)


def _validate_top_k(top_k: int) -> None:
    if top_k < 1 or not isinstance(top_k, int):
        raise InvalidArgumentException("Top k must be a positive integer.", Service.INDEX)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse


class CountItemsResponse(VectorIndexResponse):
    """Parent response type for a `count_items` request.

    Its subtypes are:
    - `CountItems.Success`
    - `CountItems.Error`

    See `PreviewVectorIndexClient` for how to work with responses.
    """


class CountItems(ABC):
    """Groups all `CountItemsResponse` derived types under a common namespace."""

    @dataclass
    class Success(CountItemsResponse):
        """Contains the result of a `count_items` request."""

        item_count: int
        """The number of items in the index."""

    class Error(CountItemsResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional

from momento_wire_types import vectorindex_pb2 as vectorindex_pb

from momento.common_data.vector_index.item import Metadata

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse
from .search import SearchHit
from .utils import pb_metadata_to_dict


class SearchAndFetchVectorsResponse(VectorIndexResponse):
    """Parent response type for a vector index `search_and_fetch_vectors` request.

    Its subtypes are:
    - `SearchAndFetchVectors.Success`
    - `SearchAndFetchVectors.Error`

    See `VectorIndexClient` for how to work with responses.
    """


class SearchAndFetchVectorsHit(SearchHit):
    def __init__(self, score: float, id: str, vector: list[float], metadata: Optional[Metadata] = None):
        """Initializes a new instance of `SearchAndFetchVectorsHit`.

        Args:
            score (float): The similarity of the hit to the query.
            id (str): The id of the hit.
            vector (list[float]): The vector of the hit.
            metadata (Optional[Metadata], optional): The metadata of the hit. Defaults to None, ie empty metadata.
        """
        super().__init__(score, id, metadata)
        self.vector = vector

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SearchAndFetchVectorsHit):
            return super().__eq__(other) and self.vector == other.vector
        return False

    def __hash__(self) -> int:
        return hash((self.score, self.id, self.vector, self.metadata))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(score={self.score!r}, id={self.id!r}, vector={self.vector!r}, metadata={self.metadata!r})"

    @staticmethod
    def from_search_hit(hit: SearchHit, vector: list[float]) -> SearchAndFetchVectorsHit:
        return SearchAndFetchVectorsHit(id=hit.id, score=hit.score, metadata=hit.metadata, vector=vector)

    @staticmethod
    def from_proto(hit: vectorindex_pb._SearchAndFetchVectorsHit) -> SearchAndFetchVectorsHit:
        metadata = pb_metadata_to_dict(hit.metadata)
        return SearchAndFetchVectorsHit(id=hit.id, score=hit.score, metadata=metadata, vector=list(hit.vector.elements))


class SearchAndFetchVectors(ABC):
    """Groups all `SearchAndFetchVectorsResponse` derived types under a common namespace."""

    @dataclass
    class Success(SearchAndFetchVectorsResponse):
        """Indicates the request was successful."""

        hits: list[SearchAndFetchVectorsHit]

    class Error(SearchAndFetchVectorsResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from enum import Enum


class SimilarityMetric(Enum):
    """The similarity metric to use when comparing vectors in the index."""

    EUCLIDEAN_SIMILARITY = "EUCLIDEAN_SIMILARITY"
    """The Euclidean distance squared between two vectors, ie the sum of squared differences between each element.
    Smaller is better. Ranges from 0 to infinity."""
    INNER_PRODUCT = "INNER_PRODUCT"
    """The inner product between two vectors, ie the sum of the element-wise products.
    Bigger is better. Ranges from 0 to infinity."""
    COSINE_SIMILARITY = "COSINE_SIMILARITY"
    """The cosine similarity between two vectors, ie the cosine of the angle between them.
    Bigger is better. Ranges from -1 to 1."""



"""Field class for filter expressions.

This class is used to create filter expressions in a more readable way::

        from momento.requests.vector_index.filters import Field

        Field("name") == "foo"
        Field("age") >= 18
        Field("tags").list_contains("books")
        (Field("year") > 2000) | (Field("year") < 1990)
"""

from __future__ import annotations

from dataclasses import dataclass

from momento_wire_types import vectorindex_pb2 as vectorindex_pb

from . import filters as F


@dataclass
class Field:
    """Represents a field in a filter expression.

    Can be used to create filter expressions in a more readable way::

        from momento.requests.vector_index.filters import Field

        Field("name") == "foo"
        Field("age") >= 18
        Field("tags").list_contains("books")
        (Field("year") > 2000) | (Field("year") < 1990)
    """

    name: str
    """The name of the field."""

    def to_filter_expression_proto(self) -> vectorindex_pb._FilterExpression:
        """Converts the field to a protobuf filter expression.

        A bare field is equivalent to the field in a boolean context.

        Returns:
            vectorindex_pb._FilterExpression: The protobuf filter expression.
        """
        return F.Equals(self.name, True).to_filter_expression_proto()

    def __eq__(self, other: str | int | float | bool) -> F.Equals:  # type: ignore
        return F.Equals(self.name, other)

    def __invert__(self) -> F.Equals:
        return F.Equals(self.name, False)

    def __ne__(self, other: str | int | float | bool) -> F.Not:  # type: ignore
        return F.Not(F.Equals(self.name, other))

    def __gt__(self, other: int | float) -> F.GreaterThan:
        return F.GreaterThan(self.name, other)

    def __ge__(self, other: int | float) -> F.GreaterThanOrEqual:
        return F.GreaterThanOrEqual(self.name, other)

    def __lt__(self, other: int | float) -> F.LessThan:
        return F.LessThan(self.name, other)

    def __le__(self, other: int | float) -> F.LessThanOrEqual:
        return F.LessThanOrEqual(self.name, other)

    def list_contains(self, value: str) -> F.ListContains:
        return F.ListContains(self.name, value)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from momento_wire_types import vectorindex_pb2 as pb

from momento.common_data.vector_index.item import Metadata
from momento.errors.exceptions import UnknownException

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse
from .utils import pb_metadata_to_dict


class GetItemMetadataBatchResponse(VectorIndexResponse):
    """Parent response type for a `get_item_metadata_batch` request.

    Its subtypes are:
    - `GetItemMetadataBatch.Success`
    - `GetItemMetadataBatch.Error`

    See `PreviewVectorIndexClient` for how to work with responses.
    """


class GetItemMetadataBatch(ABC):
    """Groups all `GetItemMetadataBatchResponse` derived types under a common namespace."""

    @dataclass
    class Success(GetItemMetadataBatchResponse):
        """Contains the result of a `get_item_metadata_batch` request."""

        values: dict[str, Metadata]
        """The metadata of the items that were found."""

        @staticmethod
        def from_proto(response: pb._GetItemMetadataBatchResponse) -> "GetItemMetadataBatch.Success":
            """Converts a proto hit to a `GetItemMetadataBatch.Success`."""
            values = {}
            for item in response.item_metadata_response:
                type = item.WhichOneof("response")
                if type == "hit":
                    values[item.hit.id] = pb_metadata_to_dict(item.hit.metadata)
                elif type == "miss":
                    pass
                else:
                    raise UnknownException(f"Unknown response type {type!r}.")

            return GetItemMetadataBatch.Success(values=values)

    class Error(GetItemMetadataBatchResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



"""Helper functions for creating gRPC channel credentials.

Note that this isn't in the _utilities init to avoid a circular import.
"""
from __future__ import annotations

import grpc

from momento.config import Configuration, VectorIndexConfiguration


def channel_credentials_from_root_certs_or_default(
    config: Configuration | VectorIndexConfiguration,
) -> grpc.ChannelCredentials:
    """Create gRPC channel credentials from the root certificates or the default credentials.

    Args:
        config (Configuration): the configuration to use.

    Returns:
        grpc.ChannelCredentials: the gRPC channel credentials.
    """
    root_certificates = config.get_transport_strategy().get_grpc_configuration().get_root_certificates_pem()
    return grpc.ssl_channel_credentials(root_certificates=root_certificates)  # type: ignore[misc]



"""Momento client library.

Instantiate a client with `CacheClient` or `CacheClientAsync` (for asyncio).
Use `CredentialProvider` to read credentials for the client.
Use `Configurations` for pre-built network configurations.
"""

import logging

from momento import logs

from .auth import CredentialProvider
from .cache_client import CacheClient
from .cache_client_async import CacheClientAsync
from .config import Configurations, TopicConfigurations, VectorIndexConfigurations
from .topic_client import TopicClient
from .topic_client_async import TopicClientAsync
from .vector_index_client import PreviewVectorIndexClient
from .vector_index_client_async import PreviewVectorIndexClientAsync

logging.getLogger("momentosdk").addHandler(logging.NullHandler())
logs.initialize_momento_logging()

__all__ = [
    "CredentialProvider",
    "Configurations",
    "TopicConfigurations",
    "VectorIndexConfigurations",
    "CacheClient",
    "CacheClientAsync",
    "TopicClient",
    "TopicClientAsync",
    "PreviewVectorIndexClient",
    "PreviewVectorIndexClientAsync",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from pathlib import Path
from typing import Optional

from momento.internal._utilities import _validate_request_timeout

from .grpc_configuration import GrpcConfiguration


class TransportStrategy(ABC):
    """Configures the network options for communicating with the Momento service."""

    @abstractmethod
    def get_grpc_configuration(self) -> GrpcConfiguration:
        """Access the gRPC configuration.

        Returns:
            GrpcConfiguration: low-level gRPC settings for the Momento client's communication.
        """
        pass

    @abstractmethod
    def with_grpc_configuration(self, grpc_configuration: GrpcConfiguration) -> TransportStrategy:
        """Copy constructor for overriding the gRPC configuration.

        Args:
            grpc_configuration (GrpcConfiguration): the new gRPC configuration.

        Returns:
            TransportStrategy: a new TransportStrategy with the specified gRPC config.
        """
        pass

    @abstractmethod
    def with_client_timeout(self, client_timeout: timedelta) -> TransportStrategy:
        """Copies the TransportStrategy and updates the copy's client-side timeout.

        Args:
            client_timeout (timedelta): the new client-side timeout.

        Returns:
            TransportStrategy: the new TransportStrategy.
        """
        pass


class StaticGrpcConfiguration(GrpcConfiguration):
    def __init__(self, deadline: timedelta, root_certificates_pem: Optional[bytes] = None):
        self._deadline = deadline
        self._root_certificates_pem = root_certificates_pem

    def get_deadline(self) -> timedelta:
        return self._deadline

    def with_deadline(self, deadline: timedelta) -> GrpcConfiguration:
        _validate_request_timeout(deadline)
        return StaticGrpcConfiguration(deadline, self._root_certificates_pem)

    def with_root_certificates_pem(self, root_certificates_pem_path: Path) -> GrpcConfiguration:
        try:
            root_certificates_pem_bytes = root_certificates_pem_path.read_bytes()
        except FileNotFoundError as e:
            raise FileNotFoundError(f"Root certificate file not found at path: {root_certificates_pem_path}") from e
        except PermissionError as e:
            raise PermissionError(f"Root certificate file not readable at path: {root_certificates_pem_path}") from e
        return StaticGrpcConfiguration(self._deadline, root_certificates_pem_bytes)

    def get_root_certificates_pem(self) -> Optional[bytes]:
        return self._root_certificates_pem


class StaticTransportStrategy(TransportStrategy):
    def __init__(self, grpc_configuration: GrpcConfiguration):
        self._grpc_configuration = grpc_configuration

    def get_grpc_configuration(self) -> GrpcConfiguration:
        return self._grpc_configuration

    def with_grpc_configuration(self, grpc_configuration: GrpcConfiguration) -> TransportStrategy:
        return StaticTransportStrategy(grpc_configuration)

    def with_client_timeout(self, client_timeout: timedelta) -> TransportStrategy:
        return self.with_grpc_configuration(self._grpc_configuration.with_deadline(client_timeout))



from __future__ import annotations

import grpc
from momento_wire_types import controlclient_pb2_grpc as control_client
from momento_wire_types import vectorindex_pb2_grpc as vector_index_client

from momento.auth import CredentialProvider
from momento.config import VectorIndexConfiguration
from momento.internal._utilities import momento_version
from momento.internal._utilities._channel_credentials import (
    channel_credentials_from_root_certs_or_default,
)

from ._add_header_client_interceptor import AddHeaderClientInterceptor, Header


class _VectorIndexControlGrpcManager:
    """Internal gRPC control manager."""

    version = momento_version

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.control_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
            interceptors=_interceptors(credential_provider.auth_token),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> control_client.ScsControlStub:
        return control_client.ScsControlStub(self._secure_channel)  # type: ignore[no-untyped-call]


class _VectorIndexDataGrpcManager:
    """Internal gRPC vector index data manager."""

    version = momento_version

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.vector_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
            interceptors=_interceptors(credential_provider.auth_token),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> vector_index_client.VectorIndexStub:
        return vector_index_client.VectorIndexStub(self._secure_channel)  # type: ignore[no-untyped-call]


def _interceptors(auth_token: str) -> list[grpc.aio.ClientInterceptor]:
    headers = [
        Header("authorization", auth_token),
        Header("agent", f"python:{_VectorIndexControlGrpcManager.version}"),
    ]
    return list(
        filter(
            None,
            [AddHeaderClientInterceptor(headers)],
        )
    )



from __future__ import annotations

from types import TracebackType
from typing import Optional, Type

from momento import logs
from momento.auth import CredentialProvider
from momento.config import VectorIndexConfiguration

try:
    from momento.internal._utilities import _validate_request_timeout
    from momento.internal.aio._vector_index_control_client import (
        _VectorIndexControlClient,
    )
    from momento.internal.aio._vector_index_data_client import _VectorIndexDataClient
except ImportError as e:
    if e.name == "cygrpc":
        import sys

        print(
            "There is an issue on M1 macs between GRPC native packaging and Python wheel tags. "
            "See https://github.com/grpc/grpc/issues/28387",
            file=sys.stderr,
        )
        print("-".join("" for _ in range(99)), file=sys.stderr)
        print("    TO WORK AROUND:", file=sys.stderr)
        print("    * Install Rosetta 2", file=sys.stderr)
        print(
            "    * Install Python from python.org (you might need to do this if you're using an arm-only build)",
            file=sys.stderr,
        )
        print("    * re-run with:", file=sys.stderr)
        print("arch -x86_64 {} {}".format(sys.executable, *sys.argv), file=sys.stderr)
        print("-".join("" for _ in range(99)), file=sys.stderr)
    raise e

from momento.requests.vector_index import AllMetadata, FilterExpression, Item, SimilarityMetric
from momento.responses.vector_index import (
    CountItemsResponse,
    CreateIndexResponse,
    DeleteIndexResponse,
    DeleteItemBatchResponse,
    GetItemBatchResponse,
    GetItemMetadataBatchResponse,
    ListIndexesResponse,
    SearchAndFetchVectorsResponse,
    SearchResponse,
    UpsertItemBatchResponse,
)


class PreviewVectorIndexClientAsync:
    """Async Vector Index Client.

    Vector and control methods return a response object unique to each request.
    The response object is resolved to a type-safe object of one of several
    sub-types. See the documentation for each response type for details.

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+ if you're deleting a key::

        response = await client.create_index(index_name, num_dimensions)
        match response:
            case CreateIndex.Success():
                ...the index was created...
            case CreateIndex.IndexAlreadyExists():
                ... the index already exists...
            case CreateIndex.Error():
                ...there was an error trying to delete the key...

    or equivalently in earlier versions of python::

        response = await client.create_index(index_name, num_dimensions)
        if isinstance(response, CreateIndex.Success):
            ...
        case isinstance(response, CreateIndex.IndexAlreadyExists):
            ... the index already exists...
        elif isinstance(response, CreateIndex.Error):
            ...
        else:
            raise Exception("This should never happen")
    """

    def __init__(self, configuration: VectorIndexConfiguration, credential_provider: CredentialProvider):
        """Instantiate a client.

        Args:
            configuration (VectorIndexConfiguration): An object holding configuration settings for communication
                with the server.
            credential_provider (CredentialProvider): An object holding the auth token and endpoint information.

        Raises:
            IllegalArgumentException: If method arguments fail validations.
        Example::

            from momento import CredentialProvider, PreviewVectorIndexClientAsync, VectorIndexConfigurations

            configuration = VectorIndexConfigurations.Laptop.latest()
            credential_provider = CredentialProvider.from_environment_variable("MOMENTO_API_KEY")
            client = PreviewVectorIndexClientAsync(configuration, credential_provider)
        """
        _validate_request_timeout(configuration.get_transport_strategy().get_grpc_configuration().get_deadline())
        self._logger = logs.logger
        self._next_client_index = 0
        self._control_client = _VectorIndexControlClient(configuration, credential_provider)
        self._data_client = _VectorIndexDataClient(configuration, credential_provider)

    async def __aenter__(self) -> PreviewVectorIndexClientAsync:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self._control_client.close()
        await self._data_client.close()

    async def count_items(self, index_name: str) -> CountItemsResponse:
        """Gets the number of items in a vector index.

        Note that if the vector index does not exist, a `NOT_FOUND` error will be returned.

        Args:
            index_name (str): Name of the index to count the items in.

        Returns:
            CountItemsResponse: The result of a count items operation.
        """
        return await self._data_client.count_items(index_name)

    async def create_index(
        self,
        index_name: str,
        num_dimensions: int,
        similarity_metric: SimilarityMetric = SimilarityMetric.COSINE_SIMILARITY,
    ) -> CreateIndexResponse:
        """Creates a vector index if it doesn't exist.

        Remark on the choice of similarity metric:
        - Cosine similarity is appropriate for most embedding models as they tend to be optimized
            for this metric.
        - If the vectors are unit normalized, cosine similarity is equivalent to inner product.
            If your vectors are already unit normalized, you can use inner product to improve
            performance.
        - Euclidean similarity, the sum of squared differences, is appropriate for datasets where
            this metric is meaningful. For example, if the vectors represent images, and the
            embedding model is trained to optimize the euclidean distance between images, then
            euclidean similarity is appropriate.

        Args:
            index_name (str): Name of the index to be created.
            num_dimensions (int): Number of dimensions of the vectors to be indexed.
            similarity_metric (SimilarityMetric): The similarity metric to use when comparing
                vectors in the index. Defaults to SimilarityMetric.COSINE_SIMILARITY.

        Returns:
            CreateIndexResponse: The result of a create index operation.
        """
        return await self._control_client.create_index(index_name, num_dimensions, similarity_metric)

    async def delete_index(self, index_name: str) -> DeleteIndexResponse:
        """Deletes a vector index and all of the items within it.

        Args:
            index_name (str): Name of the index to be deleted.

        Returns:
            DeleteIndexResponse: The result of a delete index operation.
        """
        return await self._control_client.delete_index(index_name)

    async def list_indexes(self) -> ListIndexesResponse:
        """Lists all vector indexes.

        Returns:
            ListIndexesResponse: The result of a list indexes operation.
        """
        return await self._control_client.list_indexes()

    async def upsert_item_batch(self, index_name: str, items: list[Item]) -> UpsertItemBatchResponse:
        """Upserts a batch of items into a vector index.

        If an item with the same ID already exists in the index, it will be replaced.
        Otherwise, it will be added to the index.

        Args:
            index_name (str): Name of the index to add the items into.
            items (list[Item]): The items to be added into the index.

        Returns:
            AddItemBatchResponse: The result of an add item batch operation.
        """
        return await self._data_client.upsert_item_batch(index_name, items)

    async def delete_item_batch(self, index_name: str, ids: list[str]) -> DeleteItemBatchResponse:
        """Deletes a batch of items from a vector index.

        Deletes any and all items with the given IDs from the index.

        Args:
            index_name (str): Name of the index to delete the items from.
            ids (list[str]): The IDs of the items to be deleted from the index.

        Returns:
            DeleteItemBatchResponse: The result of a delete item batch operation.
        """
        return await self._data_client.delete_item_batch(index_name, ids)

    async def search(
        self,
        index_name: str,
        query_vector: list[float],
        top_k: int = 10,
        metadata_fields: Optional[list[str]] | AllMetadata = None,
        score_threshold: Optional[float] = None,
        filter: Optional[FilterExpression] = None,
    ) -> SearchResponse:
        """Searches for the most similar vectors to the query vector in the index.

        Ranks the results using the similarity metric specified when the index was created.

        Args:
            index_name (str): Name of the index to search in.
            query_vector (list[float]): The vector to search for.
            top_k (int): The number of results to return. Defaults to 10.
            metadata_fields (Optional[list[str]] | AllMetadata): A list of metadata fields
                to return with each result. If not provided, no metadata is returned.
                If the special value `ALL_METADATA` is provided, all metadata is returned.
                Defaults to None.
            score_threshold (Optional[float]): A score threshold to filter results by.
                For cosine similarity and inner product, scores lower than the threshold
                are excluded. For euclidean similarity, scores higher than the threshold
                are excluded. The threshold is exclusive. Defaults to None, ie no threshold.
            filter (Optional[FilterExpression]): A filter expression to filter
                results by. Defaults to None, ie no filter.

        Returns:
            SearchResponse: The result of a search operation.
        """
        return await self._data_client.search(index_name, query_vector, top_k, metadata_fields, score_threshold, filter)

    async def search_and_fetch_vectors(
        self,
        index_name: str,
        query_vector: list[float],
        top_k: int = 10,
        metadata_fields: Optional[list[str]] | AllMetadata = None,
        score_threshold: Optional[float] = None,
        filter: Optional[FilterExpression] = None,
    ) -> SearchAndFetchVectorsResponse:
        """Searches for the most similar vectors to the query vector in the index.

        Ranks the results using the similarity metric specified when the index was created.
        Also returns the vectors associated with each result.

        Args:
            index_name (str): Name of the index to search in.
            query_vector (list[float]): The vector to search for.
            top_k (int): The number of results to return. Defaults to 10.
            metadata_fields (Optional[list[str]] | AllMetadata): A list of metadata fields
                to return with each result. If not provided, no metadata is returned.
                If the special value `ALL_METADATA` is provided, all metadata is returned.
                Defaults to None.
            score_threshold (Optional[float]): A score threshold to filter results by.
                For cosine similarity and inner product, scores lower than the threshold
                are excluded. For euclidean similarity, scores higher than the threshold
                are excluded. The threshold is exclusive. Defaults to None, ie no threshold.
            filter (Optional[FilterExpression]): A filter expression to filter
                results by. Defaults to None, ie no filter.

        Returns:
            SearchResponse: The result of a search operation.
        """
        return await self._data_client.search_and_fetch_vectors(
            index_name, query_vector, top_k, metadata_fields, score_threshold, filter
        )

    async def get_item_batch(self, index_name: str, ids: list[str]) -> GetItemBatchResponse:
        """Gets a batch of items from a vector index by ID.

        Args:
            index_name (str): Name of the index to get the item from.
            ids (list[str]): The IDs of the items to be retrieved from the index.

        Returns:
            GetItemBatchResponse: The result of a get item batch operation.
        """
        return await self._data_client.get_item_batch(index_name, ids)

    async def get_item_metadata_batch(self, index_name: str, ids: list[str]) -> GetItemMetadataBatchResponse:
        """Gets metadata for a batch of items from a vector index by ID.

        Args:
            index_name (str): Name of the index to get the items from.
            ids (list[str]): The IDs of the item metadata to be retrieved from the index.

        Returns:
            GetItemMetadataBatchResponse: The result of a get item metadata batch operation.
        """
        return await self._data_client.get_item_metadata_batch(index_name, ids)

    # TODO: repr



from __future__ import annotations

from abc import ABC, abstractmethod


class TopicConfigurationBase(ABC):
    @abstractmethod
    def get_max_subscriptions(self) -> int:
        pass

    @abstractmethod
    def with_max_subscriptions(self, max_subscriptions: int) -> TopicConfiguration:
        pass


class TopicConfiguration(TopicConfigurationBase):
    """Configuration options for Momento topic client."""

    def __init__(self, max_subscriptions: int = 0):
        """Instantiate a configuration.

        Args:
            max_subscriptions (int): The maximum number of subscriptions the client is expected
            to handle. Because each gRPC channel can handle 100 connections, we must explicitly
            open multiple channels to accommodate the load. NOTE: if the number of connection
            attempts exceeds the number the channels can support, program execution will block
            and hang.
        """
        self._max_subscriptions = max_subscriptions

    def get_max_subscriptions(self) -> int:
        return self._max_subscriptions

    def with_max_subscriptions(self, max_subscriptions: int) -> TopicConfiguration:
        return TopicConfiguration(max_subscriptions)



from abc import ABC

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse


class DeleteItemBatchResponse(VectorIndexResponse):
    """Parent response type for a vector index `delete_item_batch` request.

    Its subtypes are:
    - `DeleteItemBatch.Success`
    - `DeleteItemBatch.Error`

    See `VectorIndexClient` for how to work with responses.
    """


class DeleteItemBatch(ABC):
    """Groups all `DeleteItemBatchResponse` derived types under a common namespace."""

    class Success(DeleteItemBatchResponse):
        """Indicates the request was successful."""

    class Error(DeleteItemBatchResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from .topic_configuration import TopicConfiguration


class TopicConfigurations:
    class Default(TopicConfiguration):
        """Provides the recommended default configuration for topic clients."""

        @staticmethod
        def latest() -> TopicConfigurations.Default:
            """Provides the latest recommended configuration for topics.

            This configuration will be updated every time there is a new version of the laptop configuration.
            """
            return TopicConfigurations.Default.v1()

        @staticmethod
        def v1() -> TopicConfigurations.Default:
            """Provides the v1 recommended configuration for topics.

            This configuration is guaranteed not to change in future releases of the Momento Python SDK.
            """
            return TopicConfigurations.Default(max_subscriptions=0)



from abc import ABC

from momento.responses.mixins import ErrorResponseMixin
from momento.responses.response import ControlResponse


class CacheFlushResponse(ControlResponse):
    """Parent response type for a cache `flush` request.

    Its subtypes are:
    - `CacheFlush.Success`
    - `CacheFlush.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheFlush(ABC):
    """Groups all `CacheFlushResponse` derived types under a common namespace."""

    class Success(CacheFlushResponse):
        """Indicates the request was successful."""

    class Error(CacheFlushResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from datetime import timedelta
from types import TracebackType
from typing import Iterable, Optional, Type

from momento import logs
from momento.auth import CredentialProvider
from momento.config import Configuration
from momento.errors import UnknownException
from momento.requests import CollectionTtl, SortOrder
from momento.utilities.shared_sync_asyncio import (
    DEFAULT_EAGER_CONNECTION_TIMEOUT_SECONDS,
    validate_eager_connection_timeout,
)

try:
    from momento.internal._utilities import _validate_request_timeout
    from momento.internal.synchronous._scs_control_client import _ScsControlClient
    from momento.internal.synchronous._scs_data_client import _ScsDataClient
except ImportError as e:
    if e.name == "cygrpc":
        import sys

        print(
            "There is an issue on M1 macs between GRPC native packaging and Python wheel tags. "
            "See https://github.com/grpc/grpc/issues/28387",
            file=sys.stderr,
        )
        print("-".join("" for _ in range(99)), file=sys.stderr)
        print("    TO WORK AROUND:", file=sys.stderr)
        print("    * Install Rosetta 2", file=sys.stderr)
        print(
            "    * Install Python from python.org (you might need to do this if you're using an arm-only build)",
            file=sys.stderr,
        )
        print("    * re-run with:", file=sys.stderr)
        print("arch -x86_64 {} {}".format(sys.executable, *sys.argv), file=sys.stderr)
        print("-".join("" for _ in range(99)), file=sys.stderr)
    raise e

from momento.responses import (
    CacheDeleteResponse,
    CacheDictionaryFetchResponse,
    CacheDictionaryGetField,
    CacheDictionaryGetFieldResponse,
    CacheDictionaryGetFields,
    CacheDictionaryGetFieldsResponse,
    CacheDictionaryIncrementResponse,
    CacheDictionaryRemoveField,
    CacheDictionaryRemoveFieldResponse,
    CacheDictionaryRemoveFields,
    CacheDictionaryRemoveFieldsResponse,
    CacheDictionarySetField,
    CacheDictionarySetFieldResponse,
    CacheDictionarySetFields,
    CacheDictionarySetFieldsResponse,
    CacheFlushResponse,
    CacheGetResponse,
    CacheIncrementResponse,
    CacheListConcatenateBackResponse,
    CacheListConcatenateFrontResponse,
    CacheListFetchResponse,
    CacheListLengthResponse,
    CacheListPopBackResponse,
    CacheListPopFrontResponse,
    CacheListPushBackResponse,
    CacheListPushFrontResponse,
    CacheListRemoveValueResponse,
    CacheSetAddElement,
    CacheSetAddElementResponse,
    CacheSetAddElements,
    CacheSetAddElementsResponse,
    CacheSetFetchResponse,
    CacheSetIfNotExistsResponse,
    CacheSetRemoveElement,
    CacheSetRemoveElementResponse,
    CacheSetRemoveElements,
    CacheSetRemoveElementsResponse,
    CacheSetResponse,
    CacheSortedSetFetchResponse,
    CacheSortedSetGetRankResponse,
    CacheSortedSetGetScore,
    CacheSortedSetGetScoreResponse,
    CacheSortedSetGetScores,
    CacheSortedSetGetScoresResponse,
    CacheSortedSetIncrementScoreResponse,
    CacheSortedSetPutElement,
    CacheSortedSetPutElementResponse,
    CacheSortedSetPutElements,
    CacheSortedSetPutElementsResponse,
    CacheSortedSetRemoveElement,
    CacheSortedSetRemoveElementResponse,
    CacheSortedSetRemoveElements,
    CacheSortedSetRemoveElementsResponse,
    CreateCacheResponse,
    CreateSigningKeyResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    ListSigningKeysResponse,
    RevokeSigningKeyResponse,
)
from momento.typing import TDictionaryItems, TSortedSetElements


class CacheClient:
    """Synchronous Cache Client.

    Cache and control methods return a response object unique to each request.
    The response object is resolved to a type-safe object of one of several
    sub-types. See the documentation for each response type for details.

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+ if you're deleting a key::

        response = client.delete(cache_name, key)
        match response:
            case CacheDelete.Success():
                ...they key was deleted or not found...
            case CacheDelete.Error():
                ...there was an error trying to delete the key...

    or equivalently in earlier versions of python::

        response = client.delete(cache_name, key)
        if isinstance(response, CacheDelete.Success):
            ...
        elif isinstance(response, CacheDelete.Error):
            ...
        else:
            raise Exception("This should never happen")
    """

    _NUM_CLIENTS = 1
    """(async client only) For high load, we might get better performance with multiple clients,
    because the server is configured to allow a max of 100 streams per connection.
`
    In the javascript SDK, multiple clients resulted in an obvious performance improvement.
    However, in the python SDK I have not yet been able to observe a clear benefit.
    So for now, we are putting the plumbing in place so that we can more easily test performance
    with multiple connections in the future, but we are leaving the default value set to 1.

    We are hard-coding the value for now, because we haven't yet designed the API for
    users to use to configure tunables:
    https://github.com/momentohq/dev-eco-issue-tracker/issues/85
    """

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider, default_ttl: timedelta):
        """Instantiate a client.

        Args:
            configuration (Configuration): An object holding configuration settings for communication with the server.
            credential_provider (CredentialProvider): An object holding the auth token and endpoint information.
            default_ttl (timedelta): A default Time To Live timedelta for cache objects created by this client.
                It is possible to override this setting when calling the set method.

        Raises:
            IllegalArgumentException: If method arguments fail validations.
        Example::

            from datetime import timedelta
            from momento import Configurations, CredentialProvider, CacheClient

            configuration = Configurations.Laptop.latest()
            credential_provider = CredentialProvider.from_environment_variable("MOMENTO_API_KEY")
            ttl_seconds = timedelta(seconds=60)
            client = CacheClient(configuration, credential_provider, ttl_seconds)
        """
        _validate_request_timeout(configuration.get_transport_strategy().get_grpc_configuration().get_deadline())
        self._logger = logs.logger
        self._next_client_index = 0
        self._control_client = _ScsControlClient(configuration, credential_provider)
        self._cache_endpoint = credential_provider.cache_endpoint
        self._data_clients = [
            _ScsDataClient(configuration, credential_provider, default_ttl) for _ in range(CacheClient._NUM_CLIENTS)
        ]

    @staticmethod
    def create(
        configuration: Configuration,
        credential_provider: CredentialProvider,
        default_ttl: timedelta,
        eager_connection_timeout: timedelta = timedelta(seconds=DEFAULT_EAGER_CONNECTION_TIMEOUT_SECONDS),
    ) -> CacheClient:
        """Instantiate a client.

        Args:
            configuration (Configuration): An object holding configuration settings for communication with the server.
            credential_provider (CredentialProvider): An object holding the auth token and endpoint information.
            default_ttl (timedelta): A default Time To Live timedelta for cache objects created by this client.
                It is possible to override this setting when calling the set method.
            eager_connection_timeout (timedelta): An optional timeout value to eagerly connect to Momento's server.
                This helps with your client-side latencies for the initial requests. A value of 0 indicates to the
                client to not eagerly connect.

        Raises:
            IllegalArgumentException: If method arguments fail validations.
        Example::

            from datetime import timedelta
            from momento import Configurations, CredentialProvider, CacheClient

            configuration = Configurations.Laptop.latest()
            credential_provider = CredentialProvider.from_environment_variable("MOMENTO_API_KEY")
            ttl_seconds = timedelta(seconds=60)
            eager_connection_timeout = timedelta(seconds=30)
            client = CacheClient.create(configuration, credential_provider, ttl_seconds, eager_connection_timeout)
        """
        validate_eager_connection_timeout(eager_connection_timeout)
        # an explicit 0 means that the client disabled eager connections
        if eager_connection_timeout.total_seconds() != 0:
            client = CacheClient(configuration, credential_provider, default_ttl)
            for data_client in client._data_clients:
                data_client.connect(eager_connection_timeout)
        return client

    def __enter__(self) -> CacheClient:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._control_client.close()
        for data_client in self._data_clients:
            data_client.close()

    def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a cache if it doesn't exist.

        Args:
            cache_name (str): Name of the cache to be created.

        Returns:
            CreateCacheResponse: The result of a Create Cache operation.
        """
        return self._control_client.create_cache(cache_name)

    def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all of the items within it.

        Args:
            cache_name (str): Name of the cache to be deleted.

        Returns:
            DeleteCacheResponse: The result of a Delete Cache operation.
        """
        return self._control_client.delete_cache(cache_name)

    def flush_cache(self, cache_name: str) -> CacheFlushResponse:
        """Flushes a cache and empties it of all its items.

        Args:
            cache_name (str): Name of the cache to be flushed.

        Returns:
            CacheFlushResponse: The result of a Flush Cache operation.
        """
        return self._control_client.flush(cache_name)

    def list_caches(self) -> ListCachesResponse:
        """Lists all caches.

        Returns:
            ListCachesResponse: The result of a list cache operation.
        """
        return self._control_client.list_caches()

    def create_signing_key(self, ttl: timedelta) -> CreateSigningKeyResponse:
        """Creates a Momento signing key.

        Args:
            ttl (timedelta): The key's time-to-live represented as a timedelta

        Returns:
            CreateSigningKeyResponse: The result of a Create Signing key operation.
        """
        return self._control_client.create_signing_key(ttl, self._cache_endpoint)

    def revoke_signing_key(self, key_id: str) -> RevokeSigningKeyResponse:
        """Revokes a Momento signing key, all tokens signed by which will be invalid.

        Args:
            key_id (str): The id of the Momento signing key to revoke

        Returns:
            RevokeSigningKeyResponse: The result of a Revoke Signing Key operation.
        """
        return self._control_client.revoke_signing_key(key_id)

    def list_signing_keys(self) -> ListSigningKeysResponse:
        """Lists all Momento signing keys for the provided auth token.

        Returns:
            ListSigningKeysResponse: The result of a List Signing keys operation.
        """
        return self._control_client.list_signing_keys(self._cache_endpoint)

    def increment(
        self,
        cache_name: str,
        key: str | bytes,
        amount: int = 1,
        ttl: Optional[timedelta] = None,
    ) -> CacheIncrementResponse:
        """Add to a value.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str|bytes): The key to set.
            amount (int, optional): The quantity to add to the value. Defaults to 1.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheIncrementResponse
        """
        return self._data_client.increment(cache_name, key, amount, ttl)

    def set(
        self,
        cache_name: str,
        key: str | bytes,
        value: str | bytes,
        ttl: Optional[timedelta] = None,
    ) -> CacheSetResponse:
        """Set the value in cache with a given time to live (TTL) seconds.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str | bytes): The key to set.
            value (str | bytes): The value to be stored.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheSetResponse:
        """
        return self._data_client.set(cache_name, key, value, ttl)

    def set_if_not_exists(
        self,
        cache_name: str,
        key: str | bytes,
        value: str | bytes,
        ttl: Optional[timedelta] = None,
    ) -> CacheSetIfNotExistsResponse:
        """Like `set`, but it will only set if the key does not already exist.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str | bytes): The key to set.
            value (str | bytes): The value to be stored if the key does not exist.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheSetIfNotExistsResponse:
        """
        return self._data_client.set_if_not_exists(cache_name, key, value, ttl)

    def get(self, cache_name: str, key: str | bytes) -> CacheGetResponse:
        """Get the cache value stored for the given key.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            key (str | bytes): The key to lookup.

        Returns:
            CacheGetResponse:
        """
        return self._data_client.get(cache_name, key)

    def delete(self, cache_name: str, key: str | bytes) -> CacheDeleteResponse:
        """Remove the key from the cache.

        Args:
            cache_name (str): Name of the cache to delete the key from.
            key (str | bytes): The key to delete.

        Returns:
            CacheDeleteResponse:
        """
        return self._data_client.delete(cache_name, key)

    # DICTIONARY COLLECTION METHODS
    def dictionary_fetch(self, cache_name: str, dictionary_name: str) -> CacheDictionaryFetchResponse:
        """Fetch the entire dictionary from the cache.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The name of the dictionary to fetch.

        Returns:
            CacheDictionaryFetchResponse: result of the fetch operation and the associated dictionary.
        """
        return self._data_client.dictionary_fetch(cache_name, dictionary_name)

    def dictionary_get_field(
        self, cache_name: str, dictionary_name: str, field: str | bytes
    ) -> CacheDictionaryGetFieldResponse:
        """Get the cache value stored for the given dictionary and field.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The dictionary to lookup.
            field (str | bytes): The field in the dictionary to lookup.

        Returns:
            CacheDictionaryGetFieldResponse: the status and value for the field.
        """
        get_fields_response = self.dictionary_get_fields(cache_name, dictionary_name, [field])
        if isinstance(get_fields_response, CacheDictionaryGetFields.Hit):
            if len(get_fields_response.responses) == 0:
                return CacheDictionaryGetField.Error(UnknownException("Unknown get fields response had no data"))
            return get_fields_response.responses[0]
        elif isinstance(get_fields_response, CacheDictionaryGetFields.Miss):
            return CacheDictionaryGetField.Miss()
        elif isinstance(get_fields_response, CacheDictionaryGetFields.Error):
            return CacheDictionaryGetField.Error(get_fields_response.inner_exception)
        else:
            return CacheDictionaryGetField.Error(UnknownException(f"Unknown get field response: {get_fields_response}"))

    def dictionary_get_fields(
        self, cache_name: str, dictionary_name: str, fields: Iterable[str | bytes]
    ) -> CacheDictionaryGetFieldsResponse:
        """Get several values from a dictionary.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The dictionary to lookup.
            fields (Iterable[str | bytes]): The fields in the dictionary to lookup.

        Returns:
            CacheDictionaryGetFieldsResponse: the status and associated value for each field.
        """
        return self._data_client.dictionary_get_fields(cache_name, dictionary_name, fields)

    def dictionary_increment(
        self,
        cache_name: str,
        dictionary_name: str,
        field: str | bytes,
        amount: int = 1,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionaryIncrementResponse:
        """Add an integer quantity to a dictionary value.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): Name of the dictionary to increment in.
            field (str | bytes): The field to increment.
            amount (int, optional): The quantity to add to the value. May be positive, negative, or zero. Defaults to 1.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache. This TTL takes precedence over the TTL
                used when initializing a cache client. Defaults to client TTL.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionaryIncrementResponse: result of the increment operation.
        """
        return self._data_client.dictionary_increment(cache_name, dictionary_name, field, amount, ttl)

    def dictionary_remove_field(
        self, cache_name: str, dictionary_name: str, field: str | bytes
    ) -> CacheDictionaryRemoveFieldResponse:
        """Remove a field from a dictionary.

        Performs a no-op if the dictionary or field do not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to remove the field from.
            field (str | bytes): Name of the field to remove from the dictionary.

        Returns:
            CacheDictionaryRemoveFieldResponse: result of the remove operation.
        """
        remove_fields_response = self.dictionary_remove_fields(cache_name, dictionary_name, fields=[field])
        if isinstance(remove_fields_response, CacheDictionaryRemoveFields.Success):
            return CacheDictionaryRemoveField.Success()
        elif isinstance(remove_fields_response, CacheDictionaryRemoveFields.Error):
            return CacheDictionaryRemoveField.Error(remove_fields_response.inner_exception)
        else:
            return CacheDictionaryRemoveField.Error(
                UnknownException(f"Unknown remove fields response: {remove_fields_response}")
            )

    def dictionary_remove_fields(
        self, cache_name: str, dictionary_name: str, fields: Iterable[str | bytes]
    ) -> CacheDictionaryRemoveFieldsResponse:
        """Remove fields from a dictionary.

        Performs a no-op if the dictionary or a particular field does not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to remove the fields from.
            fields (Iterable[str | bytes]): The fields to remove from the dictionary.

        Returns:
            CacheDictionaryRemoveFieldsResponse: result of the remove fields operation.
        """
        return self._data_client.dictionary_remove_fields(cache_name, dictionary_name, fields)

    def dictionary_set_field(
        self,
        cache_name: str,
        dictionary_name: str,
        field: str | bytes,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionarySetFieldResponse:
        """Set the dictionary field to a value with a given time to live (TTL) seconds.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): Name of the dictionary to set.
            field (str | bytes): The field in the dictionary to set.
            value (str | bytes): The value to be stored.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache.
                This TTL takes precedence over the TTL used when initializing a cache client.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionarySetFieldResponse: result of the set operation.
        """
        set_fields_response = self.dictionary_set_fields(cache_name, dictionary_name, items={field: value}, ttl=ttl)
        if isinstance(set_fields_response, CacheDictionarySetFields.Success):
            return CacheDictionarySetField.Success()
        elif isinstance(set_fields_response, CacheDictionarySetFields.Error):
            return CacheDictionarySetField.Error(set_fields_response.inner_exception)
        else:
            return CacheDictionarySetField.Error(
                UnknownException(f"Unknown set fields response: {set_fields_response}")
            )

    def dictionary_set_fields(
        self,
        cache_name: str,
        dictionary_name: str,
        items: TDictionaryItems,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionarySetFieldsResponse:
        """Set several dictionary field-value pairs in the cache.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to set.
            items (TDictionaryItems): Field value pairs to store.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache.
                This TTL takes precedence over the TTL used when initializing a cache client.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionarySetFieldsResponse: result of the set fields operation.
        """
        return self._data_client.dictionary_set_fields(cache_name, dictionary_name, items, ttl)

    # LIST COLLECTION METHODS
    def list_concatenate_back(
        self,
        cache_name: str,
        list_name: str,
        values: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListConcatenateBackResponse:
        """Add values to the end of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to concatenate.
            values: (Iterable[str | bytes]): The values to concatenate.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_front_to_size (Optional[int]): If the list exceeds this size, remove values from
            the start of the list.

        Returns:
            CacheListConcatenateBackResponse:
        """
        return self._data_client.list_concatenate_back(cache_name, list_name, values, ttl, truncate_front_to_size)

    def list_concatenate_front(
        self,
        cache_name: str,
        list_name: str,
        values: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListConcatenateFrontResponse:
        """Add values to the start of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to concatenate.
            values: (Iterable[str | bytes]): The values to concatenate.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_back_to_size (Optional[int], optional): If the list exceeds this size, remove values from
                the end of the list.

        Returns:
            CacheListConcatenateFrontResponse:
        """
        return self._data_client.list_concatenate_front(cache_name, list_name, values, ttl, truncate_back_to_size)

    def list_fetch(self, cache_name: str, list_name: str) -> CacheListFetchResponse:
        """Gets all values from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListFetchResponse:
        """
        return self._data_client.list_fetch(cache_name, list_name)

    def list_length(self, cache_name: str, list_name: str) -> CacheListLengthResponse:
        """Gets the number of values in the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListLengthResponse:
        """
        return self._data_client.list_length(cache_name, list_name)

    def list_pop_back(self, cache_name: str, list_name: str) -> CacheListPopBackResponse:
        """Gets, removes, and returns the last value from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListPopBackResponse:
        """
        return self._data_client.list_pop_back(cache_name, list_name)

    def list_pop_front(self, cache_name: str, list_name: str) -> CacheListPopFrontResponse:
        """Gets, removes, and returns the first value from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListPopFrontResponse:
        """
        return self._data_client.list_pop_front(cache_name, list_name)

    def list_push_back(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListPushBackResponse:
        """Add values to the end of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to push to.
            value: (str | bytes): The value to push.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_front_to_size (Optional[int]): If the list exceeds this size, remove values from
            the start of the list.

        Returns:
            CacheListPushBackResponse:
        """
        return self._data_client.list_push_back(cache_name, list_name, value, ttl, truncate_front_to_size)

    def list_push_front(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListPushFrontResponse:
        """Add values to the start of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to push to.
            value: (str | bytes): The value to push.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_back_to_size (Optional[int]): If the list exceeds this size, remove values from
                the end of the list.

        Returns:
            CacheListPushFrontResponse:
        """
        return self._data_client.list_push_front(cache_name, list_name, value, ttl, truncate_back_to_size)

    def list_remove_value(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
    ) -> CacheListRemoveValueResponse:
        """Removes all matching values from the list.

        Example:
            client.list_concatenate_front(cache_name, list_name, ['up', 'up', 'down', 'down', 'left', 'right'])
            client.list_remove_value(cache_name, list_name, 'up')
            fetch_resp = client.list_fetch(cache_name, list_name)

            # ['down', 'down', 'left', 'right']
            print(fetch_resp.values_string)

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to remove values from.
            value: (str | bytes): The value to remove.

        Returns:
            CacheListRemoveValueResponse
        """
        return self._data_client.list_remove_value(cache_name, list_name, value)

    # SET COLLECTION METHODS
    def set_add_element(
        self,
        cache_name: str,
        set_name: str,
        element: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSetAddElementResponse:
        """Adds an element to a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to add to.
            element (str | bytes): The element to add.
            ttl: (CollectionTtl, optional): How to treat the set's TTL. Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSetAddElementResponse
        """
        resp = self.set_add_elements(cache_name, set_name, (element,), ttl=ttl)

        if isinstance(resp, CacheSetAddElements.Success):
            return CacheSetAddElement.Success()
        elif isinstance(resp, CacheSetAddElements.Error):
            return CacheSetAddElement.Error(resp.inner_exception)
        else:
            raise UnknownException(f"Unknown response type: {type(resp)}")

    def set_add_elements(
        self,
        cache_name: str,
        set_name: str,
        elements: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSetAddElementsResponse:
        """Add an elements to a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to add to.
            elements (Iterable[str | bytes]): The element to add.
            ttl: (CollectionTtl, optional): How to treat the set's TTL. Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSetAddElementsResponse
        """
        return self._data_client.set_add_elements(cache_name, set_name, elements, ttl)

    def set_fetch(
        self,
        cache_name: str,
        set_name: str,
    ) -> CacheSetFetchResponse:
        """Fetches a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to fetch.

        Returns:
            CacheSetFetchResponse
        """
        return self._data_client.set_fetch(cache_name, set_name)

    def set_remove_element(self, cache_name: str, set_name: str, element: str | bytes) -> CacheSetRemoveElementResponse:
        """Remove an element from a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to remove from.
            element (str | bytes): The element to remove.

        Returns:
            CacheSetRemoveElementResponse
        """
        resp = self.set_remove_elements(cache_name, set_name, (element,))
        if isinstance(resp, CacheSetRemoveElements.Success):
            return CacheSetRemoveElement.Success()
        elif isinstance(resp, CacheSetRemoveElements.Error):
            return CacheSetRemoveElement.Error(resp.inner_exception)
        else:
            raise UnknownException(f"Unknown response type: {type(resp)}")

    def set_remove_elements(
        self, cache_name: str, set_name: str, elements: Iterable[str | bytes]
    ) -> CacheSetRemoveElementsResponse:
        """Remove elements from a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to remove from.
            elements (Iterable[str | bytes]): The element to remove.

        Returns:
            CacheSetRemoveElementsResponse
        """
        return self._data_client.set_remove_elements(cache_name, set_name, elements)

    def sorted_set_put_element(
        self,
        cache_name: str,
        sorted_set_name: str,
        value: str | bytes,
        score: float,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSortedSetPutElementResponse:
        """Sets a sorted set element to a value and score with a given time to live (TTL) seconds.

        Args:
            cache_name (str): Name of the cache containing the sorted set.
            sorted_set_name (str): Name of the sorted set to set.
            value (str | bytes): The value of the element to add.
            score (float): The score of the element to add.
            ttl: (CollectionTtl, optional): How to treat the sorted set's TTL.
                Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSortedSetPutElementResponse: result of the put operation.
        """
        put_elements_response = self.sorted_set_put_elements(
            cache_name, sorted_set_name, elements={value: score}, ttl=ttl
        )
        if isinstance(put_elements_response, CacheSortedSetPutElements.Success):
            return CacheSortedSetPutElement.Success()
        elif isinstance(put_elements_response, CacheSortedSetPutElements.Error):
            return CacheSortedSetPutElement.Error(put_elements_response.inner_exception)
        else:
            return CacheSortedSetPutElement.Error(
                UnknownException(f"Unknown put elements response: {put_elements_response}")
            )

    def sorted_set_put_elements(
        self,
        cache_name: str,
        sorted_set_name: str,
        elements: TSortedSetElements,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSortedSetPutElementsResponse:
        """Puts elements into a sorted set.

        Args:
            cache_name (str): Name of the cache containing the sorted set.
            sorted_set_name (str): The name of the sorted set to add to.
            elements (TSortedSetElements): The elements to add.
            ttl: (CollectionTtl, optional): How to treat the sorted set's TTL.
                Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSortedSetPutElementsResponse
        """
        return self._data_client.sorted_set_put_elements(cache_name, sorted_set_name, elements, ttl)

    def sorted_set_fetch_by_score(
        self,
        cache_name: str,
        sorted_set_name: str,
        min_score: Optional[float] = None,
        max_score: Optional[float] = None,
        sort_order: SortOrder = SortOrder.ASCENDING,
        offset: Optional[int] = None,
        count: Optional[int] = None,
    ) -> CacheSortedSetFetchResponse:
        """Fetches a sorted set by score range.

        Args:
            cache_name (str): Name of the cache containing the sorted set.
            sorted_set_name (str): The name of the sorted set to fetch.
            min_score (Optional[float]): The minimum score of the range to fetch from the sorted set.
                                         If None, fetches from the lowest score. Defaults to None.
            max_score (Optional[float]): The maximum score of the range to fetch from the sorted set.
                                         If None, fetches until the highest score. Defaults to None.
            sort_order (SortOrder): The sort order to use when fetching the sorted set.
                                    Defaults to SortOrder.ASCENDING.
            offset (Optional[int]): The number of elements to skip before starting to return elements.
                                    If None, starts from the first element within the specified range. Defaults to None.
            count (Optional[int]): The maximum number of elements to return.
                                   If None, returns all elements within the specified range. Defaults to None.

        Returns:
            CacheSortedSetFetchResponse: The fetched sorted set data with associated metadata.
        """
        return self._data_client.sorted_set_fetch_by_score(
            cache_name, sorted_set_name, min_score, max_score, sort_order, offset, count
        )

    def sorted_set_fetch_by_rank(
        self,
        cache_name: str,
        sorted_set_name: str,
        start_rank: Optional[int] = None,
        end_rank: Optional[int] = None,
        sort_order: SortOrder = SortOrder.ASCENDING,
    ) -> CacheSortedSetFetchResponse:
        """Fetches a sorted set by rank range.

        Args:
            cache_name (str): Name of the cache containing the sorted set.
            sorted_set_name (str): The name of the sorted set to fetch.
            start_rank (Optional[int]): The start rank of the range to fetch from the sorted set.
                                        If None, fetches from the beginning of the set. Defaults to None.
            end_rank (Optional[int]): The end rank of the range to fetch from the sorted set.
                                      If None, fetches until the end of the set. Defaults to None.
            sort_order (SortOrder): The sort order to use when fetching the sorted set.
                                              Defaults to SortOrder.ASCENDING.

        Returns:
            CacheSortedSetFetchResponse: The fetched sorted set data with associated metadata.
        """
        return self._data_client.sorted_set_fetch_by_rank(cache_name, sorted_set_name, start_rank, end_rank, sort_order)

    def sorted_set_get_score(
        self, cache_name: str, sorted_set_name: str, value: str | bytes
    ) -> CacheSortedSetGetScoreResponse:
        """Get the score stored for the given sorted set and value.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): The sorted set to look up.
            value (str | bytes): The value in the sorted set to look up.

        Returns:
            CacheSortedSetGetScoreResponse: the status and score for the value.
        """
        get_scores_response = self.sorted_set_get_scores(cache_name, sorted_set_name, [value])
        if isinstance(get_scores_response, CacheSortedSetGetScores.Hit):
            if len(get_scores_response.responses) == 0:
                return CacheSortedSetGetScore.Error(UnknownException("Unknown get scores response had no data"))
            return get_scores_response.responses[0]
        elif isinstance(get_scores_response, CacheSortedSetGetScores.Miss):
            bytes_value = value.encode("utf-8") if isinstance(value, str) else value
            return CacheSortedSetGetScore.Miss(bytes_value)
        elif isinstance(get_scores_response, CacheSortedSetGetScores.Error):
            return CacheSortedSetGetScore.Error(get_scores_response.inner_exception)
        else:
            return CacheSortedSetGetScore.Error(UnknownException(f"Unknown get scores response: {get_scores_response}"))

    def sorted_set_get_scores(
        self, cache_name: str, sorted_set_name: str, values: Iterable[str | bytes]
    ) -> CacheSortedSetGetScoresResponse:
        """Get several scores from a sorted set.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): The sorted set to look up.
            values (Iterable[str | bytes]): The values in the sorted set to look up.

        Returns:
            CacheSortedSetGetScoresResponse: the status and associated score for each value.
        """
        return self._data_client.sorted_set_get_scores(cache_name, sorted_set_name, values)

    def sorted_set_get_rank(
        self, cache_name: str, sorted_set_name: str, value: str | bytes, sort_order: SortOrder = SortOrder.ASCENDING
    ) -> CacheSortedSetGetRankResponse:
        """Get the rank for the given sorted set and value.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): The sorted set to look up.
            value (str | bytes): The value in the sorted set to look up.
            sort_order (SortOrder): The sort order to use when determining the rank
                                    of the value in the sorted set. Defaults to SortOrder.ASCENDING.

        Returns:
            CacheSortedSetGetScoresResponse: the status and associated score for each value.
        """
        return self._data_client.sorted_set_get_rank(cache_name, sorted_set_name, value, sort_order)

    def sorted_set_remove_element(
        self, cache_name: str, sorted_set_name: str, value: str | bytes
    ) -> CacheSortedSetRemoveElementResponse:
        """Remove an element from a sorted set.

        Performs a no-op if the sorted set or element does not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): Name of the sorted set to remove the element from.
            value (str | bytes): Value of the element to remove from the sorted set.

        Returns:
            CacheSortedSetRemoveElementResponse: result of the remove operation.
        """
        remove_elements_response = self.sorted_set_remove_elements(cache_name, sorted_set_name, values=[value])
        if isinstance(remove_elements_response, CacheSortedSetRemoveElements.Success):
            return CacheSortedSetRemoveElement.Success()
        elif isinstance(remove_elements_response, CacheSortedSetRemoveElements.Error):
            return CacheSortedSetRemoveElement.Error(remove_elements_response.inner_exception)
        else:
            return CacheSortedSetRemoveElement.Error(
                UnknownException(f"Unknown remove elements response: {remove_elements_response}")
            )

    def sorted_set_remove_elements(
        self, cache_name: str, sorted_set_name: str, values: Iterable[str | bytes]
    ) -> CacheSortedSetRemoveElementsResponse:
        """Remove elements from a sorted set.

        Performs a no-op if the sorted or a particular element does not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): Name of the sorted set to remove the elements from.
            values (Iterable[str | bytes]): The values of the elements to remove from the sorted set.

        Returns:
            CacheSortedSetRemoveElementsResponse: result of the remove elements operation.
        """
        return self._data_client.sorted_set_remove_elements(cache_name, sorted_set_name, values)

    def sorted_set_increment_score(
        self,
        cache_name: str,
        sorted_set_name: str,
        value: str | bytes,
        score: float = 1.0,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSortedSetIncrementScoreResponse:
        """Increments the score for the given sorted set and value.

        If the value doesn't exist in the sorted set, it will be added and its score set to the given score.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            sorted_set_name (str): The sorted set to look up.
            value (str | bytes): The value in the sorted set to look up.
            score (float): The quantity to add to the score. May be positive, negative, or zero. Defaults to 1.0.
            ttl: (CollectionTtl, optional): How to treat the sorted set's TTL.
                Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSortedSetIncrementScoreResponse: the status and associated score for each value.
        """
        return self._data_client.sorted_set_increment_score(cache_name, sorted_set_name, value, score, ttl)

    @property
    def _data_client(self) -> _ScsDataClient:
        client = self._data_clients[self._next_client_index]
        self._next_client_index = (self._next_client_index + 1) % len(self._data_clients)
        return client



"""Momento network configuration module."""

from .configuration import Configuration
from .configurations import Configurations
from .topic_configuration import TopicConfiguration
from .topic_configurations import TopicConfigurations
from .vector_index_configuration import VectorIndexConfiguration
from .vector_index_configurations import VectorIndexConfigurations

__all__ = [
    "Configuration",
    "Configurations",
    "TopicConfiguration",
    "TopicConfigurations",
    "VectorIndexConfiguration",
    "VectorIndexConfigurations",
]



from momento.common_data.vector_index.item import Item

from .filter_field import Field
from .filters import FilterExpression
from .search import ALL_METADATA, AllMetadata
from .similarity_metric import SimilarityMetric

__all__ = ["Item", "AllMetadata", "ALL_METADATA", "Field", "FilterExpression", "SimilarityMetric"]



from __future__ import annotations

import asyncio
from typing import Optional

import grpc
from momento_wire_types import cacheclient_pb2_grpc as cache_client
from momento_wire_types import cachepubsub_pb2_grpc as pubsub_client
from momento_wire_types import controlclient_pb2_grpc as control_client

from momento.auth import CredentialProvider
from momento.config import Configuration, TopicConfiguration
from momento.internal._utilities import momento_version
from momento.internal._utilities._channel_credentials import (
    channel_credentials_from_root_certs_or_default,
)
from momento.retry import RetryStrategy

from ... import logs
from ._add_header_client_interceptor import (
    AddHeaderClientInterceptor,
    AddHeaderStreamingClientInterceptor,
    Header,
)
from ._retry_interceptor import RetryInterceptor


class _ControlGrpcManager:
    """Internal gRPC control manager."""

    version = momento_version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.control_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
            interceptors=_interceptors(credential_provider.auth_token, configuration.get_retry_strategy()),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> control_client.ScsControlStub:
        return control_client.ScsControlStub(self._secure_channel)  # type: ignore[no-untyped-call]


class _DataGrpcManager:
    """Internal gRPC data manager."""

    version = momento_version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._logger = logs.logger
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=channel_credentials_from_root_certs_or_default(configuration),
            interceptors=_interceptors(credential_provider.auth_token, configuration.get_retry_strategy()),
            # Here is where you would pass override configuration to the underlying C gRPC layer.
            # However, I have tried several different tuning options here and did not see any
            # performance improvements, so sticking with the defaults for now.
            # For more info on the performance investigations:
            # https://github.com/momentohq/client-sdk-python/issues/120
            # For more info on available gRPC config options:
            # https://grpc.github.io/grpc/python/grpc.html
            # https://grpc.github.io/grpc/python/glossary.html#term-channel_arguments
            # https://github.com/grpc/grpc/blob/v1.46.x/include/grpc/impl/codegen/grpc_types.h#L140
            options=[
                # ('grpc.max_concurrent_streams', 1000),
                # ('grpc.use_local_subchannel_pool', 1),
                # (experimental.ChannelOptions.SingleThreadedUnaryStream, 1)
            ],
        )

    async def eagerly_connect(self, timeout_seconds: float) -> None:
        self._logger.debug(
            f"Attempting to create an eager connection with Momento's server within {timeout_seconds} seconds"
        )
        try:
            await asyncio.wait_for(self.wait_for_ready(), timeout_seconds)
        except Exception as error:
            self._logger.debug(f"Failed to connect to the server within the given timeout. {error}")

    async def wait_for_ready(self) -> None:
        latest_state = self._secure_channel.get_state(True)  # try_to_connect
        ready: grpc.ChannelConnectivity = grpc.ChannelConnectivity.READY
        connecting: grpc.ChannelConnectivity = grpc.ChannelConnectivity.CONNECTING
        idle: grpc.ChannelConnectivity = grpc.ChannelConnectivity.IDLE

        while latest_state != ready:
            if latest_state == idle:
                self._logger.debug("State is idle; waiting to transition to CONNECTING")
            elif latest_state == connecting:
                self._logger.debug("State transitioned to CONNECTING; waiting to get READY")
            else:
                self._logger.warn(f"Unexpected connection state: {latest_state}. while trying to eagerly connect")
                break

            # This is a gRPC callback helper that prevents us from repeatedly polling on the state
            # which is highly inefficient.
            await self._secure_channel.wait_for_state_change(latest_state)
            latest_state = self._secure_channel.get_state(False)  # no need to reconnect

        if latest_state == ready:
            self._logger.debug("Connected to Momento's server! Happy Caching!")

    async def close(self) -> None:
        self._logger.debug("Closing and tearing down gRPC channel")
        await self._secure_channel.close()

    def async_stub(self) -> cache_client.ScsStub:
        return cache_client.ScsStub(self._secure_channel)  # type: ignore[no-untyped-call]


class _PubsubGrpcManager:
    """Internal gRPC pubsub manager."""

    version = momento_version

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=_interceptors(credential_provider.auth_token, None),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> pubsub_client.PubsubStub:
        return pubsub_client.PubsubStub(self._secure_channel)  # type: ignore[no-untyped-call]


class _PubsubGrpcStreamManager:
    """Internal gRPC pubsub stream manager."""

    version = momento_version

    def __init__(self, configuration: TopicConfiguration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=_stream_interceptors(credential_provider.auth_token),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> pubsub_client.PubsubStub:
        return pubsub_client.PubsubStub(self._secure_channel)  # type: ignore[no-untyped-call]


def _interceptors(auth_token: str, retry_strategy: Optional[RetryStrategy] = None) -> list[grpc.aio.ClientInterceptor]:
    headers = [
        Header("authorization", auth_token),
        Header("agent", f"python:{_ControlGrpcManager.version}"),
    ]
    return list(
        filter(
            None,
            [
                AddHeaderClientInterceptor(headers),
                RetryInterceptor(retry_strategy) if retry_strategy else None,
            ],
        )
    )


def _stream_interceptors(auth_token: str) -> list[grpc.aio.UnaryStreamClientInterceptor]:
    headers = [
        Header("authorization", auth_token),
        Header("agent", f"python:{_PubsubGrpcStreamManager.version}"),
    ]
    return [AddHeaderStreamingClientInterceptor(headers)]



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional

from momento_wire_types import vectorindex_pb2 as vectorindex_pb

from momento.common_data.vector_index.item import Metadata

from ...mixins import ErrorResponseMixin
from ..response import VectorIndexResponse
from .utils import pb_metadata_to_dict


class SearchResponse(VectorIndexResponse):
    """Parent response type for a vector index `search` request.

    Its subtypes are:
    - `Search.Success`
    - `Search.Error`

    See `VectorIndexClient` for how to work with responses.
    """


class SearchHit:
    """A single search hit."""

    def __init__(self, score: float, id: str, metadata: Optional[Metadata] = None):
        """Initializes a new instance of `SearchHit`.

        Args:
            score (float): The similarity of the hit to the query.
            id (str): The id of the hit.
            metadata (Optional[Metadata], optional): The metadata of the hit. Defaults to None, ie empty metadata.
        """
        self.score = score
        self.id = id
        self.metadata = metadata or {}

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SearchHit):
            return self.score == other.score and self.id == other.id and self.metadata == other.metadata
        return False

    def __hash__(self) -> int:
        return hash((self.score, self.id, self.metadata))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(score={self.score!r}, id={self.id!r}, metadata={self.metadata!r})"

    @staticmethod
    def from_proto(hit: vectorindex_pb._SearchHit) -> SearchHit:
        metadata = pb_metadata_to_dict(hit.metadata)
        return SearchHit(id=hit.id, score=hit.score, metadata=metadata)


class Search(ABC):
    """Groups all `SearchResponse` derived types under a common namespace."""

    @dataclass
    class Success(SearchResponse):
        """Indicates the request was successful."""

        hits: list[SearchHit]

    class Error(SearchResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



"""Detect the version of momento installed.

This module is used to detect the version of momento installed. It is used
internally to add the version to the gRPC headers.
"""

momento_version = ""

try:
    # For python < 3.8
    import importlib_metadata

    momento_version = importlib_metadata.Distribution.from_name("momento").version  # type: ignore[no-untyped-call,misc]
except ImportError:
    # For python >= 3.8
    from importlib.metadata import version  # type: ignore[import]

    momento_version = version("momento")



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSortedSetPutElementResponse(CacheResponse):
    """Response type for a `sorted_set_put_element` request.

    Its subtypes are:
    - `CacheSortedSetPutElement.Success`
    - `CacheSortedSetPutElement.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetPutElement(ABC):
    """Groups all `CacheSortedSetPutElementResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheSortedSetPutElementResponse):
        """Indicates the put was successful."""

    class Error(CacheSortedSetPutElementResponse, ErrorResponseMixin):
        """Indicates an error occurred in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['momento',
 'momento.auth',
 'momento.common_data',
 'momento.common_data.vector_index',
 'momento.config',
 'momento.config.transport',
 'momento.errors',
 'momento.internal',
 'momento.internal._utilities',
 'momento.internal.aio',
 'momento.internal.synchronous',
 'momento.requests',
 'momento.requests.vector_index',
 'momento.responses',
 'momento.responses.control',
 'momento.responses.control.cache',
 'momento.responses.control.signing_key',
 'momento.responses.data',
 'momento.responses.data.dictionary',
 'momento.responses.data.list',
 'momento.responses.data.scalar',
 'momento.responses.data.set',
 'momento.responses.data.sorted_set',
 'momento.responses.pubsub',
 'momento.responses.vector_index',
 'momento.responses.vector_index.control',
 'momento.responses.vector_index.data',
 'momento.retry',
 'momento.utilities']

package_data = \
{'': ['*']}

install_requires = \
['grpcio>=1.46.0,<2.0.0',
 'momento-wire-types>=0.105.3,<0.106.0',
 'pyjwt>=2.4.0,<3.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4']}

setup_kwargs = {
    'name': 'momento',
    'version': '1.18.0',
    'description': 'SDK for Momento',
    'long_description': '<head>\n  <meta name="Momento Python Client Library Documentation" content="Python client software development kit for Momento Cache">\n</head>\n<img src="https://docs.momentohq.com/img/momento-logo-forest.svg" alt="logo" width="400"/>\n\n[![project status](https://momentohq.github.io/standards-and-practices/badges/project-status-official.svg)](https://github.com/momentohq/standards-and-practices/blob/main/docs/momento-on-github.md)\n[![project stability](https://momentohq.github.io/standards-and-practices/badges/project-stability-stable.svg)](https://github.com/momentohq/standards-and-practices/blob/main/docs/momento-on-github.md)\n\n# Momento Python Client Library\n\nMomento Cache is a fast, simple, pay-as-you-go caching solution without any of the operational overhead\nrequired by traditional caching solutions.  This repo contains the source code for the Momento Python client library.\n\nTo get started with Momento you will need a Momento Auth Token. You can get one from the [Momento Console](https://console.gomomento.com).\n\n* Website: [https://www.gomomento.com/](https://www.gomomento.com/)\n* Momento Documentation: [https://docs.momentohq.com/](https://docs.momentohq.com/)\n* Getting Started: [https://docs.momentohq.com/getting-started](https://docs.momentohq.com/getting-started)\n* Python SDK Documentation: [https://docs.momentohq.com/develop/sdks/python](https://docs.momentohq.com/develop/sdks/python)\n* Discuss: [Momento Discord](https://discord.gg/3HkAKjUZGq)\n\n## Packages\n\nThe Momento Python SDK package is available on pypi: [momento](https://pypi.org/project/momento/).\n\n## Usage\n\nThe examples below require an environment variable named `MOMENTO_API_KEY` which must\nbe set to a valid Momento API key. You can get one from the [Momento Console](https://console.gomomento.com).\n\nPython 3.10 introduced the `match` statement, which allows for [structural pattern matching on objects](https://peps.python.org/pep-0636/#adding-a-ui-matching-objects).\nIf you are running python 3.10 or greater, here is a quickstart you can use in your own project:\n\n```python\nfrom datetime import timedelta\n\nfrom momento import CacheClient, Configurations, CredentialProvider\nfrom momento.responses import CacheGet\n\ncache_client = CacheClient(\n    Configurations.Laptop.v1(), CredentialProvider.from_environment_variable("MOMENTO_API_KEY"), timedelta(seconds=60)\n)\n\ncache_client.create_cache("cache")\ncache_client.set("cache", "my-key", "my-value")\nget_response = cache_client.get("cache", "my-key")\nmatch get_response:\n    case CacheGet.Hit() as hit:\n        print(f"Got value: {hit.value_string}")\n    case _:\n        print(f"Response was not a hit: {get_response}")\n\n```\n\nThe above code uses [structural pattern matching](https://peps.python.org/pep-0636/), a feature introduced in Python 3.10.\nUsing a Python version less than 3.10? No problem. Here is the same example compatible across all versions of Python:\n\n```python\nfrom datetime import timedelta\n\nfrom momento import CacheClient, Configurations, CredentialProvider\nfrom momento.responses import CacheGet\n\ncache_client = CacheClient(\n    configuration=Configurations.Laptop.v1(),\n    credential_provider=CredentialProvider.from_environment_variable("MOMENTO_API_KEY"),\n    default_ttl=timedelta(seconds=60),\n)\ncache_client.create_cache("cache")\ncache_client.set("cache", "myKey", "myValue")\nget_response = cache_client.get("cache", "myKey")\nif isinstance(get_response, CacheGet.Hit):\n    print(f"Got value: {get_response.value_string}")\n\n```\n\n## Getting Started and Documentation\n\nDocumentation is available on the [Momento Docs website](https://docs.momentohq.com).\n\n## Examples\n\nWorking example projects, with all required build configuration files, are available for both Python 3.10 and up\nand Python versions before 3.10:\n\n* [Python 3.10+ examples](./examples/py310)\n* [Pre-3.10 Python examples](./examples/prepy310)\n\n## Developing\n\nIf you are interested in contributing to the SDK, please see the [CONTRIBUTING](./CONTRIBUTING.md) docs.\n\n----------------------------------------------------------------------------------------\nFor more info, visit our website at [https://gomomento.com](https://gomomento.com)!\n',
    'author': 'Momento',
    'author_email': 'hello@momentohq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gomomento.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSortedSetGetRankResponse(CacheResponse):
    """Parent response type for a cache `sorted_set_get_rank` request.

    Its subtypes are:
    - `CacheSortedSetGetRank.Hit`
    - `CacheSortedSetGetRank.Miss`
    - `CacheSortedSetGetRank.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSortedSetGetRank(ABC):
    """Groups all `CacheSortedSetGetRankResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheSortedSetGetRankResponse):
        """Contains the result of a cache hit."""

        rank: float
        """The rank of the sorted set value that was queried."""

    @dataclass
    class Miss(CacheSortedSetGetRankResponse):
        """Indicates the sorted set or sorted set element does not exist."""

    class Error(CacheSortedSetGetRankResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `message`: a detailed error message.
        """
